from unittest import TestCase
from .utils import make_assembler, parse_instr_string


class TestInstrPositioning(TestCase):

    def assertRotationsEqual(self, r1, r2):
        from mccode_antlr.instr.orientation import Rotation, Matrix
        self.assertTrue(isinstance(r1, Rotation))
        self.assertTrue(isinstance(r2, Rotation))
        d = r1 - r2
        self.assertTrue(isinstance(d, Matrix))
        self.assertEqual(round(abs(d), 14), Matrix())

    def _positioning_evaluator(self, instr):
        """Common positioning checks for `test_read_positioning` and `test_assemble_positioning`"""
        from mccode_antlr.common import Expr
        from mccode_antlr.instr.orientation import Vector, Rotation
        z, o = Expr.float(0), Expr.float(1)

        left = instr.get_component('left')
        v3 = Vector(z, z, o)
        self.assertEqual(v3, left.orientation.position())
        self.assertEqual(Rotation(z, z, -o, z, o, z, o, z, z), left.orientation.rotation())
        self.assertEqual(left.orientation.rotation(), left.orientation.rotation('coordinates').inverse())

        up = instr.get_component('up')
        v4 = Vector(o, z, z)
        self.assertEqual(v4, up.orientation.position() - left.orientation.position())
        up_orientation = Rotation(z, z, -o, -o, z, z, z, o, z)
        self.assertRotationsEqual(up_orientation, up.orientation.rotation())
        self.assertEqual(Vector(o, z, o), up.orientation.position())
        self.assertRotationsEqual(up_orientation, up.orientation.rotation('coordinates').inverse())

        last = instr.get_component('last')
        v5 = Vector(z, o, z)
        self.assertRotationsEqual(up_orientation, last.orientation.rotation())
        self.assertEqual(v5, last.orientation.position() - up.orientation.position())

    def test_assemble_positioning(self):
        """Equivalent test to `test_read_positioning` but using an assembled instrument"""
        assembler = make_assembler('orientation_test')
        origin = assembler.component("origin", "Progress_bar", at=[0, 0, 0])
        left = assembler.component('left', 'Arm', at=([0, 0, 1], origin), rotate=[0, 90, 0])
        up = assembler.component("up", 'Arm', at=([0, 0, 1], left), rotate=[-90, 0, 0])
        assembler.component("last", 'Arm', at=([0, 0, 1], up), rotate=[0, 0, 0])
        self._positioning_evaluator(assembler.instrument)

    def test_read_positioning(self):
        """Equivalent test to `test_assemble_positioning` but using a parsed instrument"""
        instr_source = """DEFINE INSTRUMENT orientation_test() TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT left = Arm() AT (0, 0, 1) RELATIVE PREVIOUS ROTATED (0, 90, 0) RELATIVE origin
        COMPONENT up = Arm() AT (0, 0, 1) RELATIVE left ROTATED (-90, 0, 0) RELATIVE left
        COMPONENT last = Arm() AT (0, 0, 1) RELATIVE up
        END """
        self._positioning_evaluator(parse_instr_string(instr_source))

    def _simple_position_tests(self, instr, positions: dict):
        from mccode_antlr.common import Expr
        from mccode_antlr.instr.orientation import Vector
        positions = {k: Vector(*[Expr.float(x) for x in v]) for k, v in positions.items()}
        for instance in instr.components:
            self.assertEqual(positions[instance.name], instance.orientation.position())
            self.assertEqual(positions[instance.name], instance.orientation.position('axes'))
            self.assertEqual(positions[instance.name], instance.orientation.position('coordinates'))

    def test_simple_positioning(self):

        from math import pi, cos, sin
        instr_source = """DEFINE INSTRUMENT simple_test() TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT slit = Arm() at (0, 0, 10) RELATIVE origin
        COMPONENT sample = Arm() at (0, 0, 10) RELATIVE slit
        END"""
        instr = parse_instr_string(instr_source)
        positions = {'origin': (0, 0, 0), 'slit': (0, 0, 10), 'sample': (0, 0, 20)}
        self._simple_position_tests(instr, positions)

        instr_source = """DEFINE INSTRUMENT slightly_more_complex() TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT guide_start = Arm() AT (0.01277, 0, 1.930338) RELATIVE origin ROTATED (0, -0.56, 0) RELATIVE origin
        COMPONENT guide = Arm() AT (0, 0, 0) RELATIVE guide_start
        COMPONENT guide_end = Arm() AT (0, 0, 4.33) RELATIVE guide
        END
        """
        instr = parse_instr_string(instr_source)
        positions = {'origin': (0, 0, 0), 'guide_start': (0.01277, 0, 1.930338), 'guide': (0.01277, 0, 1.930338),
                     'guide_end': (0.01277 - 4.33*sin(pi/180*0.56), 0, 1.930338 + 4.33*cos(pi/180*0.56))}
        self._simple_position_tests(instr, positions)
        pos_hat = (0.006615, 0, 0.999978)
        pos = instr.components[2].orientation.position()
        distance = pos.length()
        vector = pos / distance
        self.assertAlmostEqual(vector.x, pos_hat[0], 6)
        self.assertAlmostEqual(vector.y, pos_hat[1], 6)
        self.assertAlmostEqual(vector.z, pos_hat[2], 6)
        self.assertAlmostEqual(distance, 1.930380239005777, 6)

        # Combining the position and rotation (reduced) operations should yield the same position
        comb = instr.components[2].orientation.combine().reduce()
        self.assertEqual(pos, comb.position())


import confuse

# Try and simplify handling configuration values need for, e.g., compiling different versions of the runtimes
# under different operating systems, while allowing a user to 'easily' override defaults if necessary.
#
# Any platform independent configuration settings can go in 'config_default.yaml'
config = confuse.LazyConfig('mccode_antlr', __name__)

# use environment variables specified as 'MCCODE_XYZ' as configuration entries 'xyz'
config.set_env()


# Add platform specific values from the local config_platforms.yaml
def _platform_defaults():
    import platform
    import yaml
    from importlib.resources import files, as_file

    platforms_file = files(__name__).joinpath('platforms.yaml')
    if not platforms_file.is_file():
        raise RuntimeError(f"Can not locate platforms.yaml in module files (looking for {platforms_file}")

    with as_file(platforms_file) as file:
        with open(file, 'r') as data:
            platform_configs = yaml.safe_load(data)

    system = platform.system()  # "Linux", "Darwin", "Windows", ...
    if system not in platform_configs:
        raise RuntimeError(f"Unknown system '{system}' for default configuration")

    return platform_configs[system]


def _common_defaults():
    import yaml
    from importlib.resources import files, as_file

    common_file = files(__name__).joinpath('config_default.yaml')
    if not common_file.is_file():
        raise RuntimeError(f"Can not locate config_default.yaml in module files (looking for {common_file})")
    with as_file(common_file) as file:
        with open(file, 'r') as data:
            common_configs = yaml.safe_load(data)

    return common_configs


def registry_defaults(libc_registry, projects: list[str]):
    """Add values for @MCCODE_*@ expansion based on the in-use LIBC registry"""
    from datetime import datetime, timezone

    def version_macro():
        from packaging.version import Version, InvalidVersion
        try:
            v = Version(libc_registry.version)
            return (v.major * 100 + v.minor) * 1000 + v.micro
        except InvalidVersion:
            return 399999  # 3.99.999 to match McCode default (MCVERSION not set)

    configs = {project: {} for project in projects}
    for project in configs.values():
        project['string'] = str(libc_registry).replace('\n', '')
        project['version'] = libc_registry.version
        project['version_macro'] = version_macro()
        project['date'] = datetime.now(timezone.utc).strftime('%Y-%m-%d')

    # Override any configurations in case this is called more than once with different information
    config.set(configs)


# By using the 'add' method, we set these as the *lowest* priority. Any user/system files will override:
config.add(_platform_defaults())
#
config.add(_common_defaults())



from pathlib import Path
from typing import Union
from loguru import logger

VERSION_NAME_KEY = 'mccode-antlr_version_data-type-name'


def _split_version_name(version_name):
    return version_name.split('/', 1)


def _write_header(group, data_type):
    from mccode_antlr.version import version
    group.attrs[VERSION_NAME_KEY] = f'{version()}/{data_type.__name__}'


def _check_header(group, data_type):
    from mccode_antlr.version import version as module_version
    if VERSION_NAME_KEY not in group.attrs:
        raise RuntimeError(f"File does not have type information")
    version, name = _split_version_name(group.attrs[VERSION_NAME_KEY])
    if version != module_version():
        raise RuntimeError(f"File was created with mccode_antlr {version}, but asked to read {module_version()}")
    if name != data_type.__name__:
        raise RuntimeError(f"Group contains mccode_antlr type {name}, but asked to read type {data_type}")


def _standard_read(typename, group, attrs, optional, required, **kwargs):
    _check_header(group, typename)
    values = {name: group.attrs.get(name) for name in attrs}
    for name in optional:
        if name in group:
            values[name] = HDF5IO.load(group[name], **kwargs)
    for name in required:
        if name in group:
            values[name] = HDF5IO.load(group[name], **kwargs)
        else:
            logger.warning('Missing required value!')
            values[name] = None
    if any(values[r] is None for r in required):
        raise ValueError(f"Could not load required values for {typename}")
    return values


def _standard_save(typename, group, data, attrs, fields, **kwargs):
    _write_header(group, typename)
    for name in attrs:
        if getattr(data, name) is not None:
            group.attrs[name] = getattr(data, name)
    for name in fields:
        if getattr(data, name) is not None:
            HDF5IO.save(group=group.create_group(name), data=getattr(data, name), **kwargs)


def _dataclass_io(real_type, attrs: tuple = None, optional: tuple = None, required: tuple = None):
    attrs, optional, required = (attrs or (), optional or (), required or ())

    class _DataclassIO:
        @staticmethod
        def load(group, **kwargs) -> real_type:
            return real_type(**_standard_read(real_type, group, attrs, optional, required, **kwargs))

        @staticmethod
        def save(group, data, **kwargs):
            _standard_save(real_type, group, data, attrs, optional+required, **kwargs)

    return _DataclassIO


class DataSourceIO:
    from mccode_antlr.common.metadata import DataSource

    @staticmethod
    def load(group, **kwargs) -> DataSource:
        values = _standard_read(DataSourceIO.DataSource, group, ('name', 'type_name'), (), (), **kwargs)
        return DataSourceIO.DataSource.from_type_name_and_name(values['type_name'], values['name'])

    @staticmethod
    def save(group, data, **kwargs):
        _standard_save(DataSourceIO.DataSource, group, data, ('name', 'type_name'), (), **kwargs)


class InstanceIO:
    from mccode_antlr.comp import Comp
    from mccode_antlr.instr import Instance
    attrs = ('name', 'removable', 'cpu', 'group')
    names = ('orientation', 'parameters', 'split', 'when', 'extend', 'jump', 'metadata')

    @staticmethod
    def load(group, instances: dict[str, Instance], components: dict[str, Comp], **kwargs) -> Instance:
        values = _standard_read(InstanceIO.Instance, group, InstanceIO.attrs, InstanceIO.names, required=(), **kwargs)
        # handle the special cases: type, at_relative, rotate_relative
        # type:
        type_name = group.attrs.get('type')
        if type_name not in components:
            raise ValueError(f"Unknown component type {type_name}")
        values['type'] = components[type_name]
        # at_relative tuple:
        vec = HDF5IO.load(group['at_vector'])
        ref = group.attrs.get('at_reference')
        if ref is not None and ref not in instances:
            raise ValueError(f"Unknown instance reference {ref}")
        values['at_relative'] = (vec, instances.get(ref) if ref is not None else None)
        # rotate_relative tuple:
        ang = HDF5IO.load(group['rotate_angles'])
        ref = group.attrs.get('rotate_reference')
        if ref is not None and ref not in instances:
            raise ValueError(f"Unknown instance reference {ref}")
        values['rotate_relative'] = (ang, instances.get(ref) if ref is not None else None)
        return InstanceIO.Instance(**values)

    @staticmethod
    def save(group, data: Instance, **kwargs):
        _standard_save(InstanceIO.Instance, group, data, InstanceIO.attrs, InstanceIO.names, **kwargs)
        # type, at_relative, rotate_relative should be _references_ to other HDF5 objects?
        # Maybe we can get away with only keeping their names, since those are unique in the instrument
        # component _type_ name:
        group.attrs['type'] = data.type.name
        # at_relative tuple:
        HDF5IO.save(group=group.create_group('at_vector'), data=data.at_relative[0])
        if data.at_relative[1] is not None:
            group.attrs['at_reference'] = data.at_relative[1].name
        # rotate_relative tuple:
        HDF5IO.save(group=group.create_group('rotate_angles'), data=data.rotate_relative[0])
        if data.rotate_relative[1] is not None:
            group.attrs['rotate_reference'] = data.rotate_relative[1].name


class InstrIO:
    from mccode_antlr.instr import Instr
    attrs = ('name', 'source')
    names = ('parameters', 'metadata', 'included', 'user', 'declare', 'initialize', 'save', 'final',
             'flags', 'registries')

    @staticmethod
    def load(group, **kwargs) -> Instr:
        values = _standard_read(InstrIO.Instr, group, InstrIO.attrs, InstrIO.names, required=(), **kwargs)
        # We had to write the component type definitions, so that they could be loaded before the instances:
        component_types = {component.name: component for component in HDF5IO.load(group['components'])}
        # And the instances need to know which component types and instances have already been loaded:
        known_instances = {}
        # Maybe we could modify the tuple loader to avoid this complexity, but this works for now
        _check_header(group['instances'], tuple)
        instance_count = group['instances'].attrs['length']
        pad = len(str(instance_count))
        instances = []
        for idx in range(instance_count):
            instance_group = group['instances'][TupleIO.entry_name(idx, pad)]
            instances.append(HDF5IO.load(instance_group, instances=known_instances, components=component_types))
            known_instances[instances[-1].name] = instances[-1]
        values['components'] = tuple(instances)
        # With all instances known, it is possible to construct the groups dictionary
        values['groups'] = HDF5IO.load(group['groups'], instances=known_instances)
        return InstrIO.Instr(**values)

    @staticmethod
    def save(group, data: Instr, **kwargs):
        _standard_save(InstrIO.Instr, group, data, InstrIO.attrs, InstrIO.names, **kwargs)
        # The groups field does not need special treatment for saving, but does for loading
        HDF5IO.save(group=group.create_group('groups'), data=data.groups)
        # the instances _must_ be recorded in order (we can recover their names later):
        HDF5IO.save(group=group.create_group('instances'), data=data.components)
        # The order of the component types in the instrument are not important.
        HDF5IO.save(group=group.create_group('components'), data=data.component_types())


class GroupIO:
    from mccode_antlr.instr import Instance, Group

    @staticmethod
    def load(group, instances: dict[str, Instance], **kwargs) -> Group:
        values = _standard_read(GroupIO.Group, group, ('name', 'index'), (), ('ids', 'member_names'), **kwargs)
        values['members'] = [instances[name] for name in values['member_names']]
        del values['member_names']
        return GroupIO.Group(**values)

    @staticmethod
    def save(group, data, **kwargs):
        _standard_save(GroupIO.Group, group, data, ('name', 'index'), ('ids', ), **kwargs)
        HDF5IO.save(group=group.create_group('member_names'), data=[member.name for member in data.members])


def RemoteRegistryIO(actual_type):
    from mccode_antlr.reader.registry import RemoteRegistry

    class _RemoteRegistryIO:
        @staticmethod
        def load(group, **kwargs) -> RemoteRegistry:
            _check_header(group, actual_type)
            values = _standard_read(actual_type, group, actual_type.file_keys(), (), (), **kwargs)
            try:
                return actual_type(**values)
            except RuntimeError:
                logger.warning(f'Unable to reconstruct {actual_type.__name__} registry from {values}')
            return RemoteRegistry('loading error', None, None, None)

        @staticmethod
        def save(group, data, **kwargs):
            _standard_save(actual_type, group, data, actual_type.file_keys(), (), **kwargs)

    return _RemoteRegistryIO


class LocalRegistryIO:
    from mccode_antlr.reader.registry import Registry, LocalRegistry

    @staticmethod
    def load(group, **kwargs) -> Registry:
        values = _standard_read(LocalRegistryIO.LocalRegistry, group, ('name', 'root'), (), (), **kwargs)
        try:
            return LocalRegistryIO.LocalRegistry(**values)
        except RuntimeError:
            logger.warning(f'Unable to reconstruct local registry from {values}')
        return LocalRegistryIO.Registry()

    @staticmethod
    def save(group, data, **kwargs):
        _standard_save(LocalRegistryIO.LocalRegistry, group, data, ('name',), (), **kwargs)
        if data.root is not None:
            group.attrs['root'] = data.root.as_posix()


def _seitz_part_io(actual_type):
    class _RotationPartIO:
        @staticmethod
        def load(group, **kwargs):
            _check_header(group, actual_type)
            v = HDF5IO.load(group['v'], **kwargs)
            if v is None:
                raise ValueError(f"Could not load v for {actual_type}")
            return actual_type(v=v)

        @staticmethod
        def save(group, data, **kwargs):
            _write_header(group, actual_type)
            if data.v is not None:
                HDF5IO.save(group=group.create_group('v'), data=data.v, **kwargs)

    return _RotationPartIO


def _named_tuple_io(typename, names):
    class _NamedTupleIO:
        @staticmethod
        def load(group, **kwargs):
            _check_header(group, typename)
            values = {name: HDF5IO.load(group[name], **kwargs) for name in names}
            if any(v is None for v in values.values()):
                raise ValueError(f"Could not load values for {typename}")
            return typename(**values)

        @staticmethod
        def save(group, data, **kwargs):
            _write_header(group, typename)
            for name in names:
                if getattr(data, name) is not None:
                    HDF5IO.save(group=group.create_group(name), data=getattr(data, name), **kwargs)

    return _NamedTupleIO


def _op_io(typename, fields: list[str]):
    class _OpIO:
        @staticmethod
        def load(group, **kwargs) -> typename:
            from mccode_antlr.common.expression import DataType
            _check_header(group, typename)
            attrs = {name: group.attrs.get(name) for name in ('op', 'style', 'data_type_name')}
            values = {name: HDF5IO.load(group[name], **kwargs) for name in fields}
            if any(v is None for v in values.values()):
                raise ValueError(f"Could not load values for {typename}")
            op = typename(attrs['op'], **values)
            op.style = attrs['style']
            if op.data_type != DataType.from_name(attrs['data_type_name']):
                raise ValueError(
                    f"Loaded {typename} with data type {op.data_type}, but expected {attrs['data_type_name']}")
            return op

        @staticmethod
        def save(group, data: typename, **kwargs):
            _write_header(group, typename)
            for name in ('op', 'style'):
                if getattr(data, name) is not None:
                    group.attrs[name] = getattr(data, name)
            group.attrs['data_type_name'] = data.data_type.name
            for name in fields:
                if getattr(data, name) is not None:
                    HDF5IO.save(group=group.create_group(name), data=getattr(data, name), **kwargs)

    return _OpIO


class ValueIO:
    from mccode_antlr.common.expression import Value

    @staticmethod
    def load(group, **kwargs) -> Value:
        from mccode_antlr.common.expression import ObjectType, DataType, ShapeType
        _check_header(group, ValueIO.Value)
        types = {f: t.from_name(group.attrs.get(f'{f}_name'))
                 for t, f in ((ObjectType, 'object_type'), (DataType, 'data_type'), (ShapeType, 'shape_type'))}
        # A missing value is valid:
        value = HDF5IO.load(group['value'], **kwargs) if 'value' in group else None
        return ValueIO.Value(**types, value=value)

    @staticmethod
    def save(group, data: Value, **kwargs):
        _write_header(group, ValueIO.Value)
        for name in ('object_type', 'data_type', 'shape_type'):
            if getattr(data, name) is not None:
                group.attrs[f'{name}_name'] = getattr(data, name).name
        if data.value is not None:
            HDF5IO.save(group=group.create_group('value'), data=data.value, **kwargs)


def _iterable_read_setup(typename, group):
    _check_header(group, typename)
    count = group.attrs['length']
    pad = len(str(count))
    return count, pad


def _iterable_save_setup(typename, group, data):
    _write_header(group, typename)
    group.attrs['length'] = len(data)
    pad = len(str(len(data)))
    return pad


class ListIO:
    @staticmethod
    def entry_name(idx, pad=2):
        return f'l{idx:0{pad}d}'

    @staticmethod
    def load(group, **kwargs) -> list:
        count, pad = _iterable_read_setup(list, group)
        return [HDF5IO.load(group=group[TupleIO.entry_name(idx, pad)], **kwargs) for idx in range(count)]

    @staticmethod
    def save(group, data: tuple, **kwargs):
        pad = _iterable_save_setup(list, group, data)
        for idx, d in enumerate(data):
            HDF5IO.save(group=group.create_group(TupleIO.entry_name(idx, pad)), data=d, **kwargs)


class TupleIO:
    @staticmethod
    def entry_name(idx, pad=2):
        return f't{idx:0{pad}d}'

    @staticmethod
    def load(group, **kwargs) -> tuple:
        count, pad = _iterable_read_setup(tuple, group)
        return tuple(HDF5IO.load(group=group[TupleIO.entry_name(idx, pad)], **kwargs) for idx in range(count))

    @staticmethod
    def save(group, data: tuple, **kwargs):
        pad = _iterable_save_setup(tuple, group, data)
        for idx, d in enumerate(data):
            HDF5IO.save(group=group.create_group(TupleIO.entry_name(idx, pad), **kwargs), data=d)


class DictIO:
    @staticmethod
    def load(group, **kwargs) -> dict:
        _check_header(group, dict)
        return {k: HDF5IO.load(group=group[k], **kwargs) for k in group}

    @staticmethod
    def save(group, data: dict, **kwargs):
        _write_header(group, dict)
        for k, v in data.items():
            HDF5IO.save(group=group.create_group(k), data=v, **kwargs)


def _direct_io(cls, convert=None):
    convert = convert or cls

    class _DirectIO:
        @staticmethod
        def load(group):
            _check_header(group, cls)
            return convert(group['entry'][()])

        @staticmethod
        def save(group, data):
            _write_header(group, cls)
            group['entry'] = data

    return _DirectIO


class HDF5IO:
    import numpy as np
    from mccode_antlr.comp import Comp
    from mccode_antlr.common import InstrumentParameter, MetaData, ComponentParameter, RawC
    from mccode_antlr.common.expression import Expr, TrinaryOp, BinaryOp, UnaryOp
    from mccode_antlr.instr.jump import Jump
    from mccode_antlr.instr.orientation import (Matrix, Vector, Angles, Rotation, Seitz, RotationX, RotationY,
                                                RotationZ, TranslationPart, Orient, Parts, Part)
    from mccode_antlr.reader.registry import ModuleRemoteRegistry, GitHubRegistry

    _handlers = {
        'Instr': InstrIO,
        'InstrumentParameter': _dataclass_io(InstrumentParameter, attrs=('name', 'unit'), required=('value',)),
        'MetaData': _dataclass_io(MetaData, attrs=('name', 'mimetype', 'value'), required=('source',)),
        'DataSource': DataSourceIO,
        'Instance': InstanceIO,
        'RawC': _dataclass_io(RawC, attrs=('filename', 'line'), required=('source',), optional=('translated',)),
        'Group': GroupIO,
        'ModuleRemoteRegistry': RemoteRegistryIO(ModuleRemoteRegistry),
        'GitHubRegistry': RemoteRegistryIO(GitHubRegistry),
        'LocalRegistry': LocalRegistryIO,
        'ComponentParameter': _dataclass_io(ComponentParameter, attrs=('name',), required=('value',)),
        'Comp': _dataclass_io(Comp, attrs=('name', 'category', 'dependency', 'acc'),
                              optional=('define', 'setting', 'output', 'metadata', 'share', 'user',
                                        'declare', 'initialize', 'trace', 'save', 'final', 'display')),
        'Orient': _dataclass_io(Orient, attrs=('_degrees',), optional=('_position', '_rotation')),
        'Parts': _dataclass_io(Parts, optional=('_stack',)),
        'Part': _dataclass_io(Part, optional=('_axes',)),
        'RotationX': _seitz_part_io(RotationX),
        'RotationY': _seitz_part_io(RotationY),
        'RotationZ': _seitz_part_io(RotationZ),
        'TranslationPart': _seitz_part_io(TranslationPart),
        'Seitz': _named_tuple_io(Seitz, ('xx', 'xy', 'xz', 'xt', 'yx', 'yy', 'yz', 'yt', 'zx', 'zy', 'zz', 'zt')),
        'Rotation': _named_tuple_io(Rotation,('xx', 'xy', 'xz', 'yx', 'yy', 'yz', 'zx', 'zy', 'zz')),
        'Matrix': _named_tuple_io(Matrix, ('xx', 'xy', 'xz', 'yx', 'yy', 'yz', 'zx', 'zy', 'zz')),
        'Vector': _named_tuple_io(Vector, ('x', 'y', 'z')),
        'Angles': _named_tuple_io(Angles, ('x', 'y', 'z')),
        'Jump': _dataclass_io(Jump, attrs=('target', 'relative_target', 'iterate', 'absolute_target'),
                              required=('condition',)),
        'Value': ValueIO,
        'Expr': _dataclass_io(Expr, optional=('expr',)),  # Not a dataclass, but sufficiently similar
        'TrinaryOp': _op_io(TrinaryOp, ['first', 'second', 'third']),
        'BinaryOp': _op_io(BinaryOp, ['left', 'right']),
        'UnaryOp': _op_io(UnaryOp, ['value']),
        'list': ListIO,
        'tuple': TupleIO,
        'dict': DictIO,
        'str': _direct_io(str, convert=lambda b: b.decode('utf-8')),
        **{t.__name__: _direct_io(t) for t in (
            int,
            np.int64,
            np.int32,
            np.uint64,
            np.uint32,
            float,
            np.float32,
            np.float64,
            bool,
            np.bool_,
            bytes,
        )}
    }

    @classmethod
    def save(cls, group, data, **kwargs):
        name = data.__class__.__name__
        if name not in cls._handlers:
            logger.warning(f'No handler for {type(data)}, skipping')
            return None
        return cls._handlers[name].save(group, data, **kwargs)

    @classmethod
    def load(cls, group, **kwargs):
        version, name = _split_version_name(group.attrs[VERSION_NAME_KEY])
        if name not in cls._handlers:
            logger.warning(f'No handler for {name}, skipping')
            return None
        return cls._handlers[name].load(group, **kwargs)


def save_hdf5(obj, filename: Union[str, Path]) -> None:
    import h5py
    with h5py.File(filename, 'w', driver='core', backing_store=True) as file:
        HDF5IO.save(file, obj)


def load_hdf5(filename: Union[str, Path]):
    import h5py
    with h5py.File(filename, 'r', driver='core', backing_store=False) as file:
        return HDF5IO.load(file)



from unittest import TestCase

class TestExtract(TestCase):
    def test_runtime_parameter(self):
        from antlr4 import InputStream
        from mccode_antlr.grammar import McInstr_parse
        from mccode_antlr.instr import InstrVisitor
        from mccode_antlr.reader import MCSTAS_REGISTRY, Reader
        from mccode_antlr.common.expression import Value, ObjectType
        from mccode_antlr.translators.c_listener import extract_c_declared_variables, evaluate_c_defined_variables
        instr_source = """
        DEFINE INSTRUMENT blah()
        DECLARE %{
            double par;
        %}
        INITIALIZE %{
            par = (1 + 2 + 3) / (7 + 8 - 9);
        %}
        TRACE
        COMPONENT origin = Progress_bar() AT (0, 0, 0) RELATIVE ABSOLUTE
        COMPONENT source = Source_gen(T1=413.5, I1=10.22e12, T2=145.8, I2=3.44e13, T3=40.1, I3=2.78e13, dist=2,
                                      radius=0.06, focus_xw=0.1, focus_yh=0.1, lambda0=1.0, dlambda=0.5)
                           AT (0, 0, 0) RELATIVE PREVIOUS
        COMPONENT monitor = PSD_monitor(nx=par, ny=par, xwidth=0.1, yheight=0.1) AT (0, 0, 10.1) RELATIVE PREVIOUS
        END
        """
        tree = McInstr_parse(InputStream(instr_source), 'prog')
        visitor = InstrVisitor(Reader(registries=[MCSTAS_REGISTRY, ]), __file__)
        # Parse the instrument definition and return an Instr object
        instr = visitor.visitProg(tree)
        nx = instr.components[-1].get_parameter('nx')

        self.assertFalse(nx.value.is_parameter)
        self.assertTrue(nx.value.is_id)

        # parse the declare block(s) to find parameter declarations
        variables = {}
        for dec in instr.declare:
            decs = extract_c_declared_variables(dec.source)
            variables.update({d.name: d.dtype for d in decs})
        # Then parse and evaluate initialize to set their value(s).
        values = {}
        for init in instr.initialize:
            values.update(evaluate_c_defined_variables(variables, init.source, verbose=False))

        self.assertEqual(values['par'], Value.float(1))




from __future__ import annotations

from pathlib import Path
from typing import Union
from mccode_antlr.instr import Instr
from mccode_antlr.reader import Registry


def parse_mccode_instr_parameters(contents: str):
    from antlr4 import InputStream
    from mccode_antlr.grammar import McInstr_parse
    from mccode_antlr.instr import InstrParametersVisitor
    visitor = InstrParametersVisitor()
    return visitor.visitProg(McInstr_parse(InputStream(contents), 'prog'))


def parse_mccode_instr(contents: str, registries: list[Registry], source: str = None) -> Instr:
    from antlr4 import InputStream
    from mccode_antlr.grammar import McInstr_parse
    from mccode_antlr.instr import InstrVisitor
    from mccode_antlr.reader import Reader
    reader = Reader(registries=registries)
    visitor = InstrVisitor(reader, source or '<string>')
    instr = visitor.visitProg(McInstr_parse(InputStream(contents), 'prog'))
    instr.flags = tuple(reader.c_flags)
    instr.registries = tuple(registries)
    return instr


def ensure_registry(needed: Registry, have: list[Registry] | None):
    if have is None or len(have) == 0:
        return [needed]
    if needed in have:
        return have
    return have + [needed]


def parse_mcstas_instr(contents: str, registries: list[Registry] | None = None) -> Instr:
    from mccode_antlr.reader import MCSTAS_REGISTRY
    return parse_mccode_instr(contents, ensure_registry(MCSTAS_REGISTRY, registries))


def parse_mcxtrace_instr(contents: str, registries: list[Registry] | None = None) -> Instr:
    from mccode_antlr.reader import MCXTRACE_REGISTRY
    return parse_mccode_instr(contents, ensure_registry(MCXTRACE_REGISTRY, registries))


def load_mccode_instr(filename: Union[str, Path], registries: list[Registry]) -> Instr:
    return parse_mccode_instr(Path(filename).read_text(), registries, source=str(filename))


def load_mcstas_instr(filename: Union[str, Path], registries: list[Registry] | None = None) -> Instr:
    from mccode_antlr.reader import MCSTAS_REGISTRY
    return load_mccode_instr(filename, ensure_registry(MCSTAS_REGISTRY, registries))


def load_mcxtrace_instr(filename: Union[str, Path], registries: list[Registry] | None = None) -> Instr:
    from mccode_antlr.reader import MCXTRACE_REGISTRY
    return load_mccode_instr(filename, ensure_registry(MCXTRACE_REGISTRY, registries))



import unittest

FAKE_COMPONENTS = dict(component_with_vector_parameter="""
DEFINE COMPONENT component_with_vector_parameter
SETTING PARAMETERS (int n, vector easy_to_spot, vector dont_set_me)
OUTPUT PARAMETERS ()
SHARE 
%{
%}
DECLARE
%{
%}
INITIALIZE
%{
  if (n < 0) {
    fprintf(stderr, "component_with_vector_parameter: n < 0\\n");
  }
  if (easy_to_spot == NULL) {
    fprintf(stderr, "component_with_vector_parameter: parameter easy_to_spot is NULL\\n");
  }
  if (dont_set_me != NULL) {
    fprintf(stderr, "component_with_vector_parameter: parameter dont_set_me is not NULL\\n");
  }
  for (int i= 0; i < n; i++) {
    printf("easy_to_spot[%d] = %g\\n", i, easy_to_spot[i]);
  }
%}
TRACE
%{
%}
FINALLY
%{
%}
MCDISPLAY
%{
%}
END
""")


INSTR_CONTENTS = dict(test_vector_parameter="""
DEFINE INSTRUMENT test_vector_parameter()
DECLARE %{
double declared_array[] = {1, 2, 3};
%}
TRACE
COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
COMPONENT one = component_with_vector_parameter(n=3, easy_to_spot=declared_array) AT (0, 0, 1) RELATIVE origin
END
""")


class ComponentVectorParameterTestCase(unittest.TestCase):
    def setUp(self):
        from mccode_antlr.assembler import Assembler
        from mccode_antlr.reader import MCSTAS_REGISTRY
        self.assembler = Assembler('test', registries=[MCSTAS_REGISTRY])
        for name, contents in FAKE_COMPONENTS.items():
            self._parse_comp(name, contents)

    def _compiler_check(self):
        import subprocess
        from mccode_antlr.config import config
        try:
            subprocess.run([config['cc'].get(str), '--version'], check=True)
        except FileNotFoundError:
            self.skipTest(f"C compiler {config['cc']} not found")

    def _compile_and_run(self, instr, parameters):
        from mccode_antlr.compiler.c import compile_instrument, CBinaryTarget, run_compiled_instrument
        from mccode_antlr.translators.target import MCSTAS_GENERATOR
        from mccode_antlr.config import config as module_config
        from tempfile import TemporaryDirectory
        from os import R_OK, access
        from pathlib import Path

        target = CBinaryTarget(mpi=False, acc=False, count=1, nexus=False)
        config = dict(default_main=True, enable_trace=False, portable=False, include_runtime=True,
                      embed_instrument_file=False, verbose=False)

        with TemporaryDirectory() as directory:
            try:
                compile_instrument(instr, target, directory, generator=MCSTAS_GENERATOR, config=config, dump_source=True)
            except RuntimeError as e:
                raise e
            binary = Path(directory).joinpath(f'{instr.name}{module_config["ext"].get(str)}')
            self.assertTrue(binary.exists())
            self.assertTrue(binary.is_file())
            self.assertTrue(access(binary, R_OK))
            result = run_compiled_instrument(binary, target, f"--dir {directory}/instr {parameters}", capture=True)
            filtered = filter(lambda x: x.startswith('easy_to_spot'), result.decode('utf8').split('\n'))
            return [float(x.split('=')[-1].strip()) for x in filtered]

    def _parse_comp(self, comp_name: str, contents: str, instance_name: str = 'instance'):
        from antlr4 import InputStream
        from mccode_antlr.grammar import McComp_parse, McComp_ErrorListener
        from mccode_antlr.comp import CompVisitor
        from mccode_antlr.reader.reader import make_reader_error_listener
        error_listener = make_reader_error_listener(
            McComp_ErrorListener, 'Component', comp_name, contents
        )
        tree = McComp_parse(InputStream(contents), 'prog', error_listener)
        # no need to call back to the reader, so we can use a dummy visitor
        visitor = CompVisitor(self.assembler.reader, __file__, instance_name=instance_name)
        res = visitor.visitProg(tree)
        self.assembler.reader.components[comp_name] = res
        return res

    def _parse_instr(self, contents: str):
        from antlr4 import InputStream
        from mccode_antlr.grammar import McInstr_parse
        from mccode_antlr.instr import InstrVisitor
        tree = McInstr_parse(InputStream(contents), 'prog')
        visitor = InstrVisitor(self.assembler.reader, 'no source')
        instr = visitor.visitProg(tree)
        instr.flags = tuple(self.assembler.reader.c_flags)
        instr.registries = tuple(self.assembler.reader.registries)
        return instr

    def _check_instr_0_properties(self, instr):
        from mccode_antlr.common.expression import DataType, ShapeType, Expr, Value
        self.assertEqual(len(instr.components), 2)
        self.assertEqual(instr.components[0].name, 'origin')
        self.assertEqual(instr.components[1].name, 'one')
        self.assertEqual(len(instr.components[1].parameters), 2)
        self.assertEqual(instr.components[1].parameters[1].name, 'easy_to_spot')
        value = instr.components[1].parameters[1].value
        self.assertEqual(value.shape_type, ShapeType.vector)
        self.assertEqual(value.data_type, DataType.float)
        self.assertTrue(value.is_vector)
        # Now weirdness
        self.assertTrue(isinstance(value, Expr))
        self.assertEqual(len(value.expr), 1)
        self.assertTrue(isinstance(value.expr[0], Value))
        #
        self.assertEqual(value.expr[0].value, "declared_array")
        # the _actual_ value is defined in the instrument declare section, we would need to compile the instrument
        # to access it -- this should be done later

    def _parse_instr_0(self):
        return self._parse_instr(INSTR_CONTENTS['test_vector_parameter'])

    def test_instr_0_parsed(self):
        self._check_instr_0_properties(self._parse_instr_0())

    def _assemble_instr_0(self):
        self.assembler.declare_array('double', 'declared_array', [1, 2, 3], source=__file__, line=1)
        self.assembler.component('origin', 'Arm', at=((0, 0, 0), 'absolute'))
        self.assembler.component('one', 'component_with_vector_parameter', at=((0, 0, 1), 'origin'),
                                 parameters=dict(n=3, easy_to_spot='declared_array'))
        return self.assembler.instrument

    def test_instr_0_assembled(self):
        self._check_instr_0_properties(self._assemble_instr_0())

    def test_parsed_instr_0_compiled(self):
        self._compiler_check()
        easy_to_spot = self._compile_and_run(self._parse_instr_0(), '')
        self.assertEqual(easy_to_spot, [1, 2, 3])

    def test_assembled_instr_0_compiled(self):
        self._compiler_check()
        easy_to_spot = self._compile_and_run(self._assemble_instr_0(), '')
        self.assertEqual(easy_to_spot, [1, 2, 3])


if __name__ == '__main__':
    unittest.main()



from .compiled import compiled, compile_and_run

@compiled
def test_parameters():
    """Ensure that instrument parameters are available for 'when' statements, e.g., in the raytrace section"""
    from mccode_antlr.loader.loader import parse_mcstas_instr
    from textwrap import dedent
    instr = parse_mcstas_instr(dedent("""\
        define instrument test_parameters(int boolean_value=1)
        declare %{int count;%}
        initialize %{count = -1;%}
        trace
        component o = Arm() at (0,0,0) absolute extend %{++count;%}
        component a = Arm() when (count%4==0) at (0,0,0) absolute 
                      extend %{printf("a=%d\\n", count); SCATTER; %}
        component b = Arm() when (count%2==0) at (0,0,0) absolute 
                      extend %{printf("b=%d\\n", count);%}
        component c = Arm() when (boolean_value && count%2) at (0,0,0) absolute
                      extend %{printf("c=%d\\n", count);%}
        end 
    """))
    n_max = 5
    for boolean_value in [0, 1]:
        results, data = compile_and_run(instr, f"-n {n_max} boolean_value={boolean_value}")
        lines = results.decode('utf-8').splitlines()
        print(lines)
        expected = []
        for i in range(n_max):
            if i % 4 == 0:
                expected.append(f'a={i}')
            if i % 2 == 0:
                expected.append(f'b={i}')
            if boolean_value and i % 2:
                expected.append(f'c={i}')

        for ex, ln in zip(expected, lines):
            assert ex == ln


@compiled
def test_declared_component_setting_parameter():
    """Ensure that instrument parameters are available for 'when' statements, e.g., in the raytrace section"""
    from mccode_antlr.loader.loader import parse_mcstas_instr
    from textwrap import dedent
    instr_source = dedent("""\
        define instrument test_declared_component_setting_parameter(dummy=1)
        declare %{
        double ex;
        double thickness;
        int values[1];
        %}
        initialize %{
        ex=0.3;
        thickness=110.0;
        values[0] = -909;
        %}
        trace
        component o = Arm() at (0,0,0) absolute
        component a = Arm() at (ex,0,0) absolute extend %{ printf("ex=%0.2f\\n", ex); %}
        component b = Progress_bar(percent=thickness) at (0,0,0) absolute
                      extend %{ printf("thickness=%0.2f\\n", thickness);%}
        component c = Al_window(thickness=-ex / values[0]) at (0,0,0) absolute
                      extend %{ printf("thickness=%0.5f\\n", thickness);%}
        end 
    """)
    instr = parse_mcstas_instr(instr_source)
    n_max = 2
    results, data = compile_and_run(instr, f"-n {n_max} dummy=2")
    lines = results.decode('utf-8').splitlines()
    print(lines)
    expected = []
    for i in range(n_max):
        expected.append(f'ex=0.30')
        expected.append(f'thickness=110.00')
        expected.append(f'thickness=0.00033')

    for ex, ln in zip(expected, lines[1:]):
        assert ex == ln


import unittest


class HDFIOTestCase(unittest.TestCase):
    def setUp(self):
        from pathlib import Path
        from tempfile import mkdtemp
        from importlib.util import find_spec
        if not find_spec('h5py'):
            self.skipTest("h5py not found -- skipping HDF5 tests")
        self.td = Path(mkdtemp())

    def tearDown(self):
        from shutil import rmtree
        rmtree(self.td)

    def test_one_axis(self):
        from mccode_antlr.io import load_hdf5, save_hdf5
        from mccode_antlr.loader import parse_mcstas_instr
        d_spacing = 3.355  # (002) for Highly-ordered Pyrolytic Graphite
        mean_energy = 5.0
        energy_width = 0.1
        instr = f"""
        DEFINE INSTRUMENT splitRunTest(a1=0, a2=0, virtual_source_x=0.05, virtual_source_y=0.1)
        TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT source = Source_simple(yheight=0.25, xwidth=0.2, dist=1.5, focus_xw=0.06, focus_yh=0.12,
                                         E0={mean_energy}, dE={energy_width})
                           AT (0, 0, 0) RELATIVE origin
        COMPONENT guide = Guide_gravity(w1 = 0.06, h1 = 0.12, w2 = 0.05, h2 = 0.1, l = 30, m = 4) 
                          AT (0, 0, 1.5) RELATIVE  PREVIOUS
        COMPONENT guide_end = Arm() WHEN (a1 == 1) AT (0, 0, 30) RELATIVE PREVIOUS
        COMPONENT aperture = Slit(xwidth=virtual_source_x, yheight=virtual_source_y) AT (0, 0, 0.01) RELATIVE PREVIOUS
        METADATA "txt" "something" %{{
            This is some unparsed metadata that will be included as a literal string in the instrument.
        %}}
        COMPONENT split_at = Arm() AT (0, 0, 0.0001) RELATIVE PREVIOUS
        SPLIT 10 COMPONENT mono_point = Arm() AT (0, 0, 0.8) RELATIVE split_at
        COMPONENT mono = Monochromator_curved(zwidth = 0.02, yheight = 0.02, NH = 13, NV = 7, DM={d_spacing}) 
                         AT (0, 0, 0) RELATIVE  mono_point ROTATED (0, a1, 0) RELATIVE mono_point
        COMPONENT sample_arm = Arm() AT (0, 0, 0) RELATIVE mono_point ROTATED (0, a2, 0) RELATIVE mono_point
        COMPONENT detector = Monitor(xwidth=0.01, yheight=0.05) AT (0, 0, 0.8) RELATIVE sample_arm
        END
        """
        instr = parse_mcstas_instr(instr)
        filename = 'test.h5'
        save_hdf5(instr, self.td.joinpath(filename))

        read_instr = load_hdf5(self.td.joinpath(filename))
        self.assertEqual(instr.name, read_instr.name)
        instance_names = [c.name for c in instr.components]
        read_instance_names = [c.name for c in read_instr.components]
        self.assertEqual(instance_names, read_instance_names)


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase
from .utils import parse_instr_string

from mccode_antlr.instr import Instr, Instance
from mccode_antlr.common.expression import Expr
from mccode_antlr.instr.orientation import Vector


class TestInstrSplit(TestCase):
    def test_mcpl_split(self):
        instr_source = """
        DEFINE INSTRUMENT test_copy()
        TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT first = Arm() AT (0, 0, 100) RELATIVE origin ROTATED (0, 90, 0) RELATIVE origin
        COMPONENT second = Arm() AT (0, 0, 1) RELATIVE first ROTATED (-90, 0, 0) RELATIVE first
        COMPONENT split_here = Arm() AT (0, 0, 0) RELATIVE second
        COMPONENT third = Arm() AT (0, 0, 10) RELATIVE second
        END
        """
        instr = parse_instr_string(instr_source)
        before, after = instr.mcpl_split('split_here', filename='test_mcpl_split')
        self.assertTrue(isinstance(before, Instr))
        self.assertTrue(isinstance(after, Instr))
        self.assertEqual(len(before.components), 4)
        self.assertEqual(len(after.components), 2)
        for inst, name in zip(before.components, ('origin', 'first', 'second', 'split_here')):
            self.assertEqual(name, inst.name)

        for inst, name in zip(after.components, ('split_here', 'third')):
            self.assertEqual(name, inst.name)

        self.assertEqual(before.components[-1].type.name, 'MCPL_output')
        self.assertEqual(after.components[0].type.name, 'MCPL_input')

        # It is imperative that the split-point did not move:
        sp = after.components[0]
        at_rel = sp.at_relative
        self.assertTrue(at_rel[1] is None)  # the position is absolute
        pos = Vector(Expr.float(1), Expr.float(0), Expr.float(100))
        self.assertEqual(pos, at_rel[0])

        # The final component had a relative position which spanned the split point, so it should now be absolute
        self.assertTrue(after.components[-1].at_relative[1] is None)
        # rounding makes third.x 1.000...07
        for a, b in zip(Vector(Expr.float(1), Expr.float(10), Expr.float(100)), after.components[-1].at_relative[0]):
            self.assertAlmostEqual(a, b)

        # The relative orientation of components 'first' and 'second' must persist into the 'before' instrument:
        first = before.get_component('first')
        self.assertEqual(first.rotate_relative[1], before.get_component('origin'))
        self.assertEqual(before.get_component('second').rotate_relative[1], first)

    def test_split_broken_reference(self):
        from textwrap import dedent
        instr = dedent("""\
        DEFINE INSTRUMENT test_tof(phase/"degree"=0)
        TRACE
        COMPONENT Origin = Progress_bar() AT (0, 0, 0) ABSOLUTE 
        COMPONENT ESS_source = ESS_butterfly(
            sector="W", beamline=4, yheight=0.03, cold_frac=0.5, dist=1, focus_xw=0.01, focus_yh=0.01, 
            c_performance=1.0, t_performance=1.0, Lmin=1, Lmax=2, tmax_multiplier=1.5, n_pulses=1, acc_power=2.0
        ) AT (0, 0, 0) ABSOLUTE 
        COMPONENT guide = Guide_gravity(w1=0.01, w2=0.05, h1=0.01, h2=0.05, l=8.0, m=3.5, G=-9.82) AT (0, 0, 1) ABSOLUTE 
        COMPONENT guide_end = Arm() AT (0, 0, 8.0) RELATIVE guide
        COMPONENT chopper = DiskChopper(radius=0.35, nu=14, phase=phase, theta_0=115) AT (0, 0, 0.01) RELATIVE guide_end
        COMPONENT split_before = Arm() AT (0, 0, 1e-08) RELATIVE chopper
        COMPONENT split_at = Arm() AT (0, 0, 0) RELATIVE split_before
        COMPONENT split_after = Arm() AT (0, 0, 0) RELATIVE split_before
        COMPONENT sample = Incoherent(
            radius=0.005, yheight=0.02, thickness=0.001, focus_ah=2.0, focus_aw=2.0, target_x=1.0, target_y=0.0, 
            target_z=0.0, Etrans=0.0, deltaE=0.2
        ) AT (0, 0, 1) RELATIVE chopper 
        COMPONENT a4 = Arm() AT (0, 0, 0) RELATIVE sample ROTATED (0, 90, 0) RELATIVE sample
        COMPONENT analyzer = Arm() AT (0, 0, 1) RELATIVE a4
        COMPONENT a6 = Arm() AT (0, 0, 0) RELATIVE analyzer ROTATED (0, -90, 0) RELATIVE analyzer
        COMPONENT detector = Arm() AT (0, 0, 1) RELATIVE a6
        END""")
        instr = parse_instr_string(instr)
        first, second = instr.split('split_at', remove_unused_parameters=True)
        # The last component in the first part of the instrument _is_ split_at
        self.assertEqual('split_at', first.components[-1].name)
        # The position of the split_at component should be absolute and the same as in the main instrument
        split_at = second.components[0]
        self.assertEqual(split_at.name, 'split_at')
        at_ref = split_at.at_relative
        self.assertTrue(isinstance(at_ref[0], Vector))
        self.assertTrue(at_ref[1] is None)
        v = Vector(Expr.float(0), Expr.float(0), Expr.float(1 + 8 + 0.01 + 1E-8))
        self.assertEqual(v, at_ref[0])

        # The sample _should not_ still depend on the chopper, which is only present in the first instrument!
        sample = second.components[2]
        self.assertEqual(sample.name, 'sample')
        at_ref = sample.at_relative
        self.assertTrue(isinstance(at_ref[0], Vector))
        self.assertFalse(isinstance(at_ref[1], Instance))
        self.assertTrue(at_ref[1] is None)
        # the position of the sample depends on the source, guide, guide-end, and chopper
        v = Vector(Expr.float(0), Expr.float(0), Expr.float(1 + 8 + 0.01 + 1))
        self.assertEqual(v, at_ref[0])

        # But everything after _should_ still depend on relative positions
        for i, name in enumerate(('a4', 'analyzer', 'a6', 'detector')):
            # Verify that we positioned the components relative in the initial instrument
            i_inst = instr.components[len(first.components)+2+i]
            self.assertEqual(name, i_inst.name)
            self.assertTrue(i_inst.at_relative[1] is not None)
            self.assertTrue(i_inst.rotate_relative[1] is not None)
            self.assertTrue(instr.has_component_named(i_inst.at_relative[1].name))
            self.assertTrue(second.has_component_named(i_inst.at_relative[1].name))
            # And that, since they're relative to second-instrument components, they're still relative
            s_inst = second.components[3+i]
            self.assertEqual(name, s_inst.name)
            self.assertTrue(s_inst.at_relative[1] is not None)
            self.assertTrue(s_inst.rotate_relative[1] is not None)



from loguru import logger
from dataclasses import dataclass, field
from typing import TypeVar, Union
from ..comp import Comp
from ..common import Expr, Value, Mode
from ..common import InstrumentParameter, ComponentParameter, MetaData, parameter_name_present, RawC, blocks_to_raw_c
from .orientation import Orient, Vector, Angles
from .jump import Jump

InstanceReference = TypeVar('InstanceReference', bound='Instance')
VectorReference = tuple[Vector, InstanceReference]
AnglesReference = tuple[Angles, InstanceReference]

@dataclass
class Instance:
    """Intermediate representation of a McCode component instance

    Read from a .instr file TRACE section, using one or more .comp sources
    For output to a runtime source file
    """
    name: str
    type: Comp
    at_relative: VectorReference
    rotate_relative: AnglesReference
    orientation: Orient = None
    parameters: tuple[ComponentParameter, ...] = field(default_factory=tuple)
    removable: bool = False
    cpu: bool = False
    split: Expr = None
    when: Expr = None
    group: str = None
    extend: tuple[RawC, ...] = field(default_factory=tuple)
    jump: tuple[Jump, ...] = field(default_factory=tuple)
    metadata: tuple[MetaData, ...] = field(default_factory=tuple)
    mode: Mode = Mode.normal

    def __repr__(self):
        return f'Instance({self.name}, {self.type.name})'

    def to_file(self, output, wrapper=None, full=True):
        if self.cpu:
            print(wrapper.line('CPU', []), file=output)

        instance_parameters = wrapper.hide(', '.join(p.to_string(wrapper=wrapper) for p in self.parameters))
        line = wrapper.bold('COMPONENT') + f' {self.name} = {self.type.name}({instance_parameters}) '

        if self.when is not None:
            line += wrapper.bold('WHEN') + ' ' + wrapper.escape(str(self.when)) + ' '

        def rf(which, x, required=False):
            absolute = wrapper.bold('ABSOLUTE')
            relative = wrapper.bold('RELATIVE')
            return _triplet_ref_str(wrapper.bold(which), x, absolute, relative, required)

        # The "AT ..." statement is required even when it is "AT (0, 0, 0) ABSOLUTE"
        line += rf('AT', self.at_relative, required=True) + ' '
        line += rf('ROTATED', self.rotate_relative) + wrapper.br()
        print(line, file=output)

        if not full:
            return  # Skip the rest of the output

        if self.group is not None:
            print(wrapper.line('GROUP', [self.group]), file=output)
        if self.extend:
            extends = '\n'.join(str(ext) for ext in self.extend)
            print(wrapper.block('EXTEND', extends), file=output)
        for jump in self.jump:
            jump.to_file(output, wrapper)
        for metadata in self.metadata:
            metadata.to_file(output, wrapper)

    def to_string(self, wrapper, full=True):
        from io import StringIO
        output = StringIO()
        self.to_file(output, wrapper=wrapper, full=full)
        return output.getvalue()

    def __str__(self):
        from mccode_antlr.common import TextWrapper
        return self.to_string(TextWrapper())

    def partial_str(self):
        from mccode_antlr.common import TextWrapper
        return self.to_string(TextWrapper(), full=False)

    @classmethod
    def from_instance(cls, name: str, ref: InstanceReference, at: VectorReference, rotate: AnglesReference):
        # from copy import deepcopy
        # copy each of: parameters, extend, group, jump, when, metadata
        return cls(name, ref.type, at, rotate,
                   parameters=tuple([par for par in ref.parameters]),
                   when=ref.when, group=ref.group,
                   extend=tuple([ext for ext in ref.extend]),
                   jump=tuple([jmp for jmp in ref.jump]),
                   metadata=tuple([md for md in ref.metadata]),
                   mode=ref.mode)

    def __post_init__(self):
        if self.mode != Mode.minimal and self.orientation is None:
            ar, rr = self.at_relative, self.rotate_relative
            if not isinstance(ar[0], Vector) or not isinstance(rr[0], Angles):
                logger.warning(f'Expected {ar=} and {rr=} to be Vector and Angles respectively')
            if rr[1] is None and ar[1] is not None:
                logger.warning(f'Expected rotation reference to be specified when at reference is specified')
            at = ar[0] if isinstance(ar[0], Vector) else Vector(*ar[0])
            an, ar = (ar[1].name, ar[1].orientation) if ar[1] else ("ABSOLUTE", None)
            rt = rr[0] if isinstance(rr[0], Angles) else Angles(*rr[0])
            rn, rr = (rr[1].name, rr[1].orientation) if rr[1] else (an, ar)
            self.orientation = Orient.from_dependent_orientations(ar, at, rr, rt)

    def set_parameter(self, name: str, value, overwrite=False, allow_repeated=True):
        if not parameter_name_present(self.type.define, name) and not parameter_name_present(self.type.setting, name):
            raise RuntimeError(f"Unknown parameter {name} for component type {self.type.name}")
        if parameter_name_present(self.parameters, name):
            if overwrite:
                self.parameters = tuple(x for x in self.parameters if name != x.name)
            elif allow_repeated:
                par = [p for p in self.parameters if name == p.name][0]
                logger.info(f'Multiple definitions of {name} in component instance {self.name}')
                if par.value != value:
                    logger.info(f'  first-encountered value {par.value} retained')
                    logger.info(f'  newly-encountered value {value} dropped')
            else:
                raise RuntimeError(f"Multiple definitions of {name} in component instance {self.name}")
        p = self.type.get_parameter(name)

        if not p.compatible_value(value):
            logger.debug(f'{p=}, {name=}, {value=}')
            raise RuntimeError(f"Provided value for parameter {name} is not compatible with {self.type.name}")

        if p.value.is_vector and isinstance(value, str):
            # FIXME can this be more general? Do we _need_ to treat vectors differently?
            value = Expr(Value(value, data_type=p.value.data_type, shape_type=p.value.shape_type))
        elif isinstance(value, str):
            value = Expr.parse(value)
        elif not isinstance(value, Expr):
            # Copy the data_type of the component definition parameter
            # -- thus if value is a str but an int or float is expected, we will know it is an identifier
            value = Expr(Value(value, data_type=p.value.data_type))

        # 2023-09-14 This did nothing. Why was this here?
        # # is this parameter value *actually* an instrument parameter *name*
        # if value.is_id:
        #     pass
        # # If a parameter is set to an instrument parameter name, we need to keep track of that here:
        # TODO: Either add a reference to the containing instrument (and carry that around always)
        #       Or perform this check when it comes time to translate the whole instrument :/

        self.parameters += (ComponentParameter(p.name, value), )

    def verify_parameters(self, instrument_parameters: tuple[InstrumentParameter, ...]):
        """Check for instance parameters which are identifiers that match instrument parameter names,
        and flag them as parameter objects"""
        instrument_parameter_names = [x.name for x in instrument_parameters]
        for par in self.parameters:
            par.value.verify_parameters(instrument_parameter_names)

    def get_parameter(self, name: str):
        for par in self.parameters:
            if par.name == name:
                return par
        return self.type.get_parameter(name)

    def defines_parameter(self, name: str):
        """Check whether this instance has defined the named parameter"""
        return parameter_name_present(self.parameters, name)

    def set_parameters(self, **kwargs):
        for name, value in kwargs.items():
            self.set_parameter(name, value)

    def REMOVABLE(self):
        self.removable = True

    def CPU(self):
        self.cpu = True

    def SPLIT(self, count):
        if isinstance(count, str):
            count = Expr.parse(count)
        if not isinstance(count, Expr):
            raise ValueError(f'Expected provided SPLIT expression to be an Expr not a {type(count)}')
        self.split = count

    def WHEN(self, expr):
        if isinstance(expr, str):
            expr = Expr.parse(expr)
        if not isinstance(expr, Expr):
            raise ValueError(f'Expected provided WHEN expression to be an Expr not a {type(expr)}')
        if expr.is_constant:
            raise RuntimeError(f'Evaluated WHEN statement {expr} would be constant at runtime!')
        self.when = expr

    def GROUP(self, name: str):
        self.group = name

    def EXTEND(self, *blocks):
        self.extend += blocks_to_raw_c(*blocks)

    def JUMP(self, *jumps):
        self.jump += jumps

    def add_metadata(self, m: MetaData):
        if any([x.name == m.name for x in self.metadata]):
            self.metadata = tuple([x for x in self.metadata if x.name != m.name])
        self.metadata += (m, )

    def collect_metadata(self):
        # A component declaration and instance can define metadata with the same name
        # When they do, the metadata from *the instance* should take precedence
        md = {m.name: m for m in self.type.collect_metadata()}
        md.update({m.name: m for m in self.metadata})
        return tuple(md.values())

    def copy(self):
        return Instance(self.name, self.type, self.at_relative, self.rotate_relative,
                        orientation=self.orientation, parameters=self.parameters,
                        removable=self.removable, cpu=self.cpu, split=self.split, when=self.when,
                        group=self.group, extend=self.extend, jump=self.jump, metadata=self.metadata)

    def parameter_used(self, name: str):
        if any([name in par.value for par in self.parameters]):
            return True
        if name in self.at_relative[0] or name in self.rotate_relative[0] or name in self.orientation:
            return True
        if name in (self.split or []) or name in (self.when or []):
            return True
        for block in self.extend:
            if name in block:
                return True
        for jump in self.jump:
            if name in jump:
                return True
        return False


def _triplet_ref_str(which, tr: Union[VectorReference, AnglesReference], absolute, relative, required=False):
    pos, ref = tr
    if isinstance(pos, tuple):
        pos = Vector(*pos)
    if pos.is_null() and ref is None and not required:
        return ''
    return f'{which} {pos} {absolute if ref is None else f"{relative} {ref.name}"}'



from mccode_antlr.translators.c_listener import (
    extract_c_declared_variables_and_defined_types as extract,
    CDeclarator, CFuncPointer
)
from textwrap import dedent

def test_alias_types():
    block = dedent("""\
    typedef unsigned int age_t;
    """)
    variables, types = extract(block)
    assert len(variables) == 0
    assert len(types) == 1
    assert types[0] == 'age_t'


def test_stackoverflow_struct_type():
    block = dedent("""\
    typedef struct nodeT {
        void *content;
        struct nodeT *next;
    } Node;
    """)
    variables, types = extract(block)
    assert len(variables) == 0
    assert len(types) == 1
    assert 'Node' == types[0]

def test_stackoverflow_enum_type():
    block = dedent("""\
    typedef enum season_t { SPRING, SUMMER, FALL, WINTER } Season;
    """)
    variables, types = extract(block)
    assert len(variables) == 0
    assert len(types) == 1
    assert 'Season' == types[0]

def test_repeated_struct_name():
    block = dedent("""\
    typedef struct PolyhedronIndexValuePair {
        int index;
        double value;
    } PolyhedronIndexValuePair;
    """)
    variables, types = extract(block)

    assert len(variables) == 0
    assert len(types) == 1
    assert types[0] == 'PolyhedronIndexValuePair'


def test_mccode_style_typedef():
    block = dedent("""\
    struct _struct_Progress_bar {
      char     _name[256]; /* e.g. instance of Progress_bar name */
      char     _type[13]; /* Progress_bar */
      long     _index; /* index in TRACE list */
      Coords   _position_absolute;
      Coords   _position_relative; /* wrt PREVIOUS */
      Rotation _rotation_absolute;
      Rotation _rotation_relative; /* wrt PREVIOUS */
      int      _rotation_is_identity;
      int      _position_relative_is_zero;
      _class_Progress_bar_parameters _parameters;
    };
    typedef struct _struct_Progress_bar _class_Progress_bar;
    """)
    variables, types = extract(block)
    assert len(variables) == 0
    assert len(types) == 1
    assert types[0] == "_class_Progress_bar"


def test_assignments():
    block = dedent("""\
    int blah=1;
    double yarg;
    // yarg = 2.0; // this parser only handles _declarations_ no out-of-declaration assignments allowed.
    char mmmm[11] = "0123456789";
    """)
    variables, types = extract(block)
    assert len(variables) == 3
    assert len(types) == 0
    expected = [
        CDeclarator(dtype='int', declare='blah', init='1'),
        CDeclarator(dtype='double', declare='yarg'),
        CDeclarator(dtype='char', declare='mmmm', init='"0123456789"', elements=11),
    ]
    for x in expected:
        assert x in variables


def test_struct_declaration():
    block = dedent("""\
    struct my_struct_type the_struct;
    struct another_struct a_struct_with_values={0, 1.0, "two"};
    struct the_third_s * ptr_to_third_struct;
    """)
    variables, types = extract(block)
    assert len(variables) == 3
    assert len(types) == 0
    expected = [
        CDeclarator(dtype='struct my_struct_type', declare='the_struct'),
        CDeclarator(dtype='struct another_struct', declare='a_struct_with_values', init='{0, 1.0, "two"}'),
        CDeclarator(dtype='struct the_third_s', declare='ptr_to_third_struct', pointer='*')
    ]
    assert all(x in variables for x in expected)


def test_typedef_declaration():
    block = dedent("""\
    typedef double blah;
    blah really_a_double=1.0f;
    blah * double_ptr=NULL;
    blah double_array[10];
    """)
    variables, types = extract(block)
    assert len(variables) == 3
    assert len(types) == 1
    assert types[0] == "blah"
    expected = [
        CDeclarator(dtype='blah', declare='really_a_double', init='1.0f'),
        CDeclarator(dtype='blah', declare='double_ptr', pointer='*', init='NULL'),
        CDeclarator(dtype='blah', declare='double_array', elements=10)
    ]
    assert all(x in variables for x in expected)


def test_flatellipse_finite_mirror():
    # A component defines a function handle (then doesn't use it).
    block = dedent("""\
        //Scene where all geometry is added to
    Scene s;
    //point structure
    Point p1;
    //Function to handle Conic-Neutron collisions with reflectivity from McStas Tables
    void traceNeutronConicWithTables(_class_particle* p, ConicSurf c);
    double *rfront_inner;//all r-distances at lStart for all mirror surfaces
    int silicon; // +1: neutron in silicon, -1: neutron in air, 0: mirrorwidth is 0; neutron cannot be in silicon and also does not track mirror transitions
    t_Table rsTable;
    """)
    variables, types = extract(block)
    assert len(types) == 0
    assert len(variables) == 6
    expected = [
        CDeclarator(dtype='Scene', declare='s'),
        CDeclarator(dtype='Point', declare='p1'),
        # the function pre-declaration _IS NOT_ a function pointer (and should be in DECLARE)
        CDeclarator(
            dtype='void',
            declare='traceNeutronConicWithTables(_class_particle* p, ConicSurf c)',
        ),
        CDeclarator(dtype='double', pointer='*', declare='rfront_inner'),
        CDeclarator(dtype='int', declare='silicon'),
        CDeclarator(dtype='t_Table', declare='rsTable')
    ]
    for x, y in zip(expected, variables):
        assert x == y


def test_function_pointer_declaration():
    block = dedent(r"""
    int (*fun_ptr)(int, int);
    int (*fun_ptr_ar3[3])(int, int);
    int (*fun_ptr_arr[])(int, int) = {add, sub, mul};
    """)
    variables, types = extract(block)
    assert len(types) == 0
    assert len(variables) == 3
    expected = [
        CDeclarator(
            dtype='int',
            declare=CFuncPointer(
                declare=CDeclarator(pointer='*', declare='fun_ptr'),
                args='int, int',
            ),
        ),
        CDeclarator(
            dtype='int',
            declare=CFuncPointer(
                declare=CDeclarator(pointer='*', declare='fun_ptr_ar3', elements=3),
                args='int, int',
            ),
        ),
        CDeclarator(
            dtype='int',
            declare=CFuncPointer(
                declare=CDeclarator(pointer='*', declare='fun_ptr_arr', elements=0),
                args='int, int',
            ),
            init='{add, sub, mul}',
        ),
    ]
    for x, y in zip(expected, variables):
        assert x == y
    members = [
        'int (* fun_ptr)(int, int)',
        'int (* fun_ptr_ar3[3])(int, int)',
        'int (* fun_ptr_arr[3])(int, int)',
    ]
    for x, y in zip(members, variables):
        assert x == y.as_struct_member()


from .instr import Instr
from .instance import Instance
from .orientation import Orient, Parts, Part, TranslationPart, RotationPart
from .jump import Jump
from .group import Group
from .visitor import InstrVisitor, InstrParametersVisitor


__all__ = [
    'Instr',
    'Instance',
    'TranslationPart',
    'RotationPart',
    'Part',
    'Parts',
    'Orient',
    'Jump',
    'Group',
    'InstrVisitor',
    'InstrParametersVisitor',
]



from .expression import (
    Expr, UnaryOp, BinaryOp, Value, ObjectType, ShapeType, DataType, TrinaryOp
)

def visitSomething(obj, ctx):
    pass

def getExpr(obj, ctx):
    return obj.visit(ctx)

# TODO (maybe) Add control and statements into McCode, requiring some form of global stack.
def visitExpressionUnaryPM(obj, ctx):
    right = obj.visit(ctx.expr())
    if isinstance(right, str):
        return '-' + right if ctx.Plus() is None else right
    return -right if ctx.Plus() is None else right

def visitExpressionGrouping(obj, ctx):
    return Expr(UnaryOp('__group__', obj.visit(ctx.expr())))

def visitExpressionFloat(obj, ctx):
    return Expr.float(str(ctx.FloatingLiteral()))

def visitExpressionPointerAccess(obj, ctx):
    pointer = Expr(Value(str(ctx.Identifier()), object_type=ObjectType.identifier))
    return Expr(BinaryOp('__pointer_access__', pointer, obj.visit(ctx.expr())))

def visitExpressionStructAccess(obj, ctx):
    struct = Expr(Value(str(ctx.Identifier()), object_type=ObjectType.identifier))
    return Expr(BinaryOp('__struct_access__', struct, obj.visit(ctx.expr())))

def visitExpressionArrayAccess(obj, ctx):
    array = Expr(Value(str(ctx.Identifier()), object_type=ObjectType.identifier, shape_type=ShapeType.vector))
    return Expr(BinaryOp('__getitem__', array, obj.visit(ctx.expr())))

def visitExpressionIdentifier(obj, ctx):
    # check if this identifier is an InstrumentParameter name:
    name = str(ctx.Identifier())
    inst_par = obj.state.get_parameter(name, None)
    obj = ObjectType.parameter if inst_par is not None else ObjectType.identifier
    dat = inst_par.value.data_type if inst_par is not None else DataType.undefined
    return Expr(Value(name, data_type=dat, object_type=obj))

def visitExpressionInteger(obj, ctx):
    return Expr.int(str(ctx.IntegerLiteral()))

def visitExpressionZero(obj, ctx):
    return Expr.int(0)

def visitExpressionExponentiation(obj, ctx):
    base, exponent = [obj.visit(ex) for ex in ctx.expr()]
    return base ** exponent

def visitExpressionBinaryPM(obj, ctx):
    # `ctx.left` and `ctx.right` are not exposed by speedy-antlr-tool
    left, right = [obj.visit(ex) for ex in ctx.expr()]
    return left + right if ctx.Minus() is None else left - right

def visitExpressionFunctionCall(obj, ctx):
    function = Value(str(ctx.Identifier()), object_type=ObjectType.function)
    # `ctx.args` isn't exposed by speedy-antlr-tool, but the function call is
    args = [obj.visit(arg).expr[0] for arg in ctx.expr()]  # each is a Value, UnaryOp, or BinaryOp?
    return Expr(BinaryOp('__call__', function, args))

def visitExpressionBinaryMD(obj, ctx):
    # `ctx.left` and `ctx.right` are not exposed by speedy-antlr-tool
    left, right = [obj.visit(ex) for ex in ctx.expr()]
    return left * right if ctx.Div() is None else left / right

def visitExpressionBinaryMod(obj, ctx):
    # `ctx.left` and `ctx.right` are not exposed by speedy-antlr-tool
    left, right = [obj.visit(ex) for ex in ctx.expr()]
    return Expr(BinaryOp('%', left, right))

def visitExpressionBinaryLeftShift(obj, ctx):
    # `ctx.left` and `ctx.right` are not exposed by speedy-antlr-tool
    left, right = [obj.visit(ex) for ex in ctx.expr()]
    return Expr(BinaryOp('<<', left, right))

def visitExpressionBinaryRightShift(obj, ctx):
    # `ctx.left` and `ctx.right` are not exposed by speedy-antlr-tool
    left, right = [obj.visit(ex) for ex in ctx.expr()]
    return Expr(BinaryOp('>>', left, right))

def visitInitializerlist(obj, ctx):
    # `ctx.values` is not exposed by speedy-antlr-tool
    values = [obj.visit(x).expr[0].value for x in ctx.expr()]
    return Expr(Value(values, object_type=ObjectType.initializer_list, shape_type=ShapeType.vector))

def visitExpressionUnaryLogic(obj, ctx):
    expr = obj.visit(ctx.expr())
    op = 'unknown'
    if ctx.Not() is not None:
        op = '__not__'
    return Expr(UnaryOp(op, expr))

def visitExpressionBinaryLogic(obj, ctx):
    # `ctx.left` and `ctx.right` are not exposed by speedy-antlr-tool
    left, right = [obj.visit(ex) for ex in ctx.expr()]
    op = 'unknown'
    if ctx.AndAnd() is not None:
        op = '__and__'
    elif ctx.OrOr() is not None:
        op = '__or__'
    return Expr(BinaryOp(op, left, right))

def visitExpressionTrinaryLogic(obj, ctx):
    # `ctx.test`, `ctx.true_`, and `ctx._false` not exposed by speedy-antlr-tool
    test, true, false = [obj.visit(x) for x in ctx.expr()]
    return Expr(TrinaryOp('__trinary__', test, true, false))

def visitExpressionBinaryEqual(obj, ctx):
    # `ctx.left` and `ctx.right` are not exposed by speedy-antlr-tool
    left, right = [obj.visit(ex) for ex in ctx.expr()]
    return Expr(BinaryOp('__eq__', left, right))

def visitExpressionBinaryLessEqual(obj, ctx):
    # `ctx.left` and `ctx.right` are not exposed by speedy-antlr-tool
    left, right = [obj.visit(ex) for ex in ctx.expr()]
    return Expr(BinaryOp('__le__', left, right))

def visitExpressionBinaryGreaterEqual(obj, ctx):
    # `ctx.left` and `ctx.right` are not exposed by speedy-antlr-tool
    left, right = [obj.visit(ex) for ex in ctx.expr()]
    return Expr(BinaryOp('__ge__', left, right))

def visitExpressionBinaryLess(obj, ctx):
    # `ctx.left` and `ctx.right` are not exposed by speedy-antlr-tool
    left, right = [obj.visit(ex) for ex in ctx.expr()]
    return Expr(BinaryOp('__lt__', left, right))

def visitExpressionBinaryGreater(obj, ctx):
    # `ctx.left` and `ctx.right` are not exposed by speedy-antlr-tool
    left, right = [obj.visit(ex) for ex in ctx.expr()]
    return Expr(BinaryOp('__gt__', left, right))

def visitExpressionString(obj, ctx):
    strings = ''.join(str(sl).strip('"') for sl in ctx.StringLiteral())
    return Expr.str(f'"{strings}"')


common_visitors = (
    ('getExpr', getExpr),
    ('visitExpressionUnaryPM', visitExpressionUnaryPM),
    ('visitExpressionGrouping', visitExpressionGrouping),
    ('visitExpressionFloat', visitExpressionFloat),
    ('visitExpressionPointerAccess', visitExpressionPointerAccess),
    ('visitExpressionStructAccess', visitExpressionStructAccess),
    ('visitExpressionArrayAccess', visitExpressionArrayAccess),
    ('visitExpressionIdentifier', visitExpressionIdentifier),
    ('visitExpressionInteger', visitExpressionInteger),
    ('visitExpressionZero', visitExpressionZero),
    ('visitExpressionExponentiation', visitExpressionExponentiation),
    ('visitExpressionBinaryPM', visitExpressionBinaryPM),
    ('visitExpressionFunctionCall', visitExpressionFunctionCall),
    ('visitExpressionBinaryMD', visitExpressionBinaryMD),
    ('visitExpressionBinaryMod', visitExpressionBinaryMod),
    ('visitExpressionBinaryLeftShift', visitExpressionBinaryLeftShift),
    ('visitExpressionBinaryRightShift', visitExpressionBinaryRightShift),
    ('visitInitializerlist', visitInitializerlist),
    ('visitExpressionUnaryLogic', visitExpressionUnaryLogic),
    ('visitExpressionBinaryLogic', visitExpressionBinaryLogic),
    ('visitExpressionTrinaryLogic', visitExpressionTrinaryLogic),
    ('visitExpressionBinaryEqual', visitExpressionBinaryEqual),
    ('visitExpressionBinaryLessEqual', visitExpressionBinaryLessEqual),
    ('visitExpressionBinaryGreaterEqual', visitExpressionBinaryGreaterEqual),
    ('visitExpressionBinaryLess', visitExpressionBinaryLess),
    ('visitExpressionBinaryGreater', visitExpressionBinaryGreater),
    ('visitExpressionString', visitExpressionString),
)

def add_common_visitors(grammar_visitor):
    for name, func in common_visitors:
        setattr(grammar_visitor, name, func)



from __future__ import annotations

import pooch
from pathlib import Path
from re import Pattern
from mccode_antlr.version import version as mccode_antlr_version


def ensure_regex_pattern(pattern):
    import re
    if not isinstance(pattern, Pattern) and isinstance(pattern, str):
        pattern = re.compile(pattern)
    if not isinstance(pattern, Pattern):
        raise RuntimeError("No valid regex pattern from input")
    return pattern


def simple_url_validator(url: str, file_ok=False):
    from urllib.parse import urlparse

    if not isinstance(url, str):
        return False
    try:
        result = urlparse(url)
    except AttributeError:
        return False
    if not result.scheme:
        return False
    if file_ok:
        if result.scheme == 'file':
            print("Constructing a RemoteRegistry for a file:// URL will likely duplicate files!")
        if result.scheme != 'file' and not result.netloc:
            return False
    elif not result.netloc:
        return False
    return True


class Registry:
    name = None
    root = None
    pooch = None
    version = None
    priority: int = 0

    def __str__(self):
        from mccode_antlr.common import TextWrapper
        return self.to_string(TextWrapper())

    def to_string(self, wrapper):
        from io import StringIO
        output = StringIO()
        self.to_file(output, wrapper)
        return output.getvalue()

    def to_file(self, output, wrapper):
        print('Registry<>', file=output)

    def known(self, name: str, ext: str = None):
        pass

    def unique(self, name: str):
        pass

    def fullname(self, name: str, ext: str = None):
        pass

    def is_available(self, name: str, ext: str = None):
        pass

    def path(self, name: str, ext: str = None) -> Path:
        pass

    def filenames(self) -> list[str]:
        pass

    def search(self, regex: Pattern):
        """Return filenames containing the regex pattern, uses regex search"""
        regex = ensure_regex_pattern(regex)
        return [x for x in self.filenames() if regex.search(x) is not None]

    def match(self, regex: Pattern):
        """Return regex *matching* registered file names -- which *start* with the regex pattern"""
        regex = ensure_regex_pattern(regex)
        return [x for x in self.filenames() if regex.match(x) is not None]

    def contents(self, *args, **kwargs):
        """Return the text contents of a Registry file"""
        return self.path(*args, **kwargs).read_text()


def _name_plus_suffix(name: str, suffix: str = None):
    path = Path(name)
    if suffix is not None and path.suffix != suffix:
        return path.with_suffix(f'{path.suffix}{suffix}').as_posix()
    return path.as_posix()


def find_registry_file(name: str):
    """Find a registry file in the mccode_antlr package"""
    from importlib.resources import files
    from importlib.metadata import distribution
    from json import loads
    if isinstance(name, Path):
        name = name.as_posix()
    if files('mccode_antlr').joinpath(name).is_file():
        return files('mccode_antlr').joinpath(name)
    info = loads(distribution('mccode_antlr').read_text('direct_url.json'))
    if 'dir_info' in info and 'editable' in info['dir_info'] and info['dir_info']['editable'] and 'url' in info:
        path = Path(info['url'].split('file://')[1]).joinpath('mccode_antlr', name)
        return path if path.is_file() else None
    return None


class RemoteRegistry(Registry):
    def __init__(self, name: str, url: str | None, version: str | None, filename: str | None, priority: int = 0):
        self.name = name
        self.url = url
        self.version = version
        self.filename = filename
        self.pooch = None
        self.priority = priority

    @classmethod
    def file_keys(cls) -> tuple[str, ...]:
        return 'name', 'url', 'version', 'filename'

    def file_contents(self) -> dict[str, str]:
        return {key: getattr(self, key) or '' for key in self.file_keys()}

    def to_file(self, output, wrapper):
        wp = wrapper.parameter
        wv = wrapper.value
        contents = '(' + ', '.join([wp(n) + '=' + wv(v) for n, v in self.file_contents().items()]) + ')'
        print(wrapper.line(self.__class__.__name__, [contents], ''), file=output)

    def known(self, name: str, ext: str = None):
        compare = _name_plus_suffix(name, ext)
        # the files *in* the registry are already posix paths, so that makes life easier
        if any(x.endswith(compare) for x in self.pooch.registry_files):
            return True
        # fall back to matching without the extension:
        if any(Path(x).with_suffix('').as_posix().endswith(compare) for x in self.pooch.registry_files):
            return True
        # Or matching *any* file that contains name
        return any(name in x for x in self.pooch.registry_files)

    def unique(self, name: str):
        return sum(name in x for x in self.pooch.registry_files) == 1

    def fullname(self, name: str, ext: str = None, exact: bool = True):
        compare = _name_plus_suffix(name, ext)
        # the files *in* the registry are already posix paths, so that makes life easier
        # first step, exact match:
        for registry_file in self.pooch.registry_files:
            if registry_file == compare:
                return registry_file
        # second step, exact match to filename:
        for registry_file in self.pooch.registry_files:
            if Path(registry_file).parts[-1] == compare:
                return registry_file
        # fall back to matching without the extension:
        if not exact:
            for registry_file in self.pooch.registry_files:
                if Path(registry_file).with_suffix('').as_posix() == compare:
                    return registry_file
            # nearly-giving-up match filename without extension against the comparison
            for registry_file in self.pooch.registry_files:
                if Path(registry_file).with_suffix('').parts[-1] == compare:
                    return registry_file
        # Or matching *any* file that contains name
        matches = list(filter(lambda x: name in x, self.pooch.registry_files))
        if len(matches) != 1:
            raise RuntimeError(f'More than one match for {name}:{ext}, which is required of:\n{matches}')
        return matches[0]

    def is_available(self, name: str, ext: str = None, exact: bool = True):
        return self.pooch.registry_files(self.fullname(name, ext, exact))

    def path(self, name: str, ext: str = None, exact: bool = True) -> Path:
        return Path(self.pooch.fetch(self.fullname(name, ext, exact)))

    def filenames(self) -> list[str]:
        return [x for x in self.pooch.registry_files]

    def __eq__(self, other):
        if not isinstance(other, Registry):
            return False
        if other.name != self.name:
            return False
        if other.pooch is None:
            return False
        if other.pooch.path != self.pooch.path:
            return False
        if other.pooch.base_url != self.pooch.base_url:
            return False
        return True


class ModuleRemoteRegistry(RemoteRegistry):
    def __init__(self, name: str, url: str, filename=None, version=None, priority: int = 0):
        super().__init__(name, url, version, filename, priority)
        self.pooch = pooch.create(
            path=pooch.os_cache(f'mccode_antlr-{self.name}'),
            base_url=self.url,
            version=self.version or mccode_antlr_version(),
            version_dev="main",
            registry=None,
        )
        if isinstance(self.filename, Path):
            self.pooch.load_registry(self.filename)
        else:
            filepath = find_registry_file(self.filename)
            if filepath is None:
                raise RuntimeError(f"Provided filename {self.filename} is not a path or file packaged with this module")
            self.pooch.load_registry(filepath)


class GitHubRegistry(RemoteRegistry):
    def __init__(self, name: str, url: str, version: str, filename: str | None = None,
                 registry: str | dict | None = None, priority: int = 0):
        if filename is None:
            filename = f'{name}-registry.txt'
        super().__init__(name, url, version, filename, priority)
        import requests
        base_url = f'{self.url}/raw/{self.version}/'
        # If registry is a string url, we expect the registry file to be available from _that_ url
        self._stashed_registry = None
        if isinstance(registry, str) and simple_url_validator(registry, file_ok=True):
            self._stashed_registry = registry
            registry = f'{registry}/raw/{self.version}/'
        # We allow a full-dictionary to be provided, otherwise we expect the registry file to be available from the
        # base_url where all subsequent files are also expected to be available
        if not isinstance(registry, dict):
            r = requests.get((registry or base_url) + (self.filename or 'pooch-registry.txt'))
            if not r.ok:
                raise RuntimeError(f"Could not retrieve {r.url} because {r.reason}")
            registry = {k: v for k, v in [x.split(maxsplit=1) for x in r.text.split('\n') if len(x)]}

        self.pooch = pooch.create(
            path=pooch.os_cache(self.name),
            base_url=base_url,
            version=version if version.startswith('v') else None,
            version_dev="main",
            registry=registry,
        )

    @property
    def registry(self):
        return self._stashed_registry

    def file_contents(self) -> dict[str, str]:
        fc = super().file_contents()
        fc['registry'] = self._stashed_registry or ''
        return fc

    @classmethod
    def file_keys(cls) -> tuple[str, ...]:
        return super().file_keys() + ('registry',)


class LocalRegistry(Registry):
    def __init__(self, name: str, root: str, priority: int = 10):
        self.name = name
        self.root = Path(root)
        self.version = mccode_antlr_version()
        self.priority = priority

    def to_file(self, output, wrapper):
        contents = '(' + ', '.join([
            wrapper.parameter('name') + '=' + wrapper.value(self.name),
            wrapper.parameter('root') + '=' + wrapper.url(self.root.as_posix()),
        ]) + ')'
        print(wrapper.line('LocalRegistry', [contents], ''), file=output)

    def _filetype_iterator(self, filetype: str):
        return self.root.glob(f'**/*.{filetype}')

    def _file_iterator(self, name: str):
        return self.root.glob(f'**/{name}')

    def _exact_file_iterator(self, name: str):
        return self.root.glob(name)

    def known(self, name: str, ext: str = None):
        compare = _name_plus_suffix(name, ext)
        return len(list(self._file_iterator(compare))) > 0

    def unique(self, name: str):
        return len(list(self._file_iterator(name))) == 1

    def fullname(self, name: str, ext: str = None, exact: bool = True):
        compare = _name_plus_suffix(name, ext)
        # Complete match
        is_compare = list(self._exact_file_iterator(compare))
        if len(is_compare) == 1:
            return is_compare[0]
        # Complete match if name happens to be missing the extension
        is_name = list(self._exact_file_iterator(name))
        if len(is_name) == 1:
            return is_name[0]
        if not exact:
            ends_with_compare = list(self._file_iterator(compare))
            if len(ends_with_compare) == 1:
                return ends_with_compare[0]
            # Complete match if name happens to be missing the extension
            ends_with_name = list(self._file_iterator(name))
            if len(ends_with_name) == 1:
                return ends_with_name[0]
        # Or matching *any* file that contains name
        matches = list(self._file_iterator(name))
        if len(matches) != 1:
            raise RuntimeError(f'More than one match for {name}:{ext}, which is required of:\n{matches}')
        return matches[0]

    def is_available(self, name: str, ext: str = None):
        return self.known(name, ext)

    def path(self, name: str, ext: str = None, exact: bool = True) -> Path:
        return self.root.joinpath(self.fullname(name, ext, exact))

    def filenames(self) -> list[str]:
        return [str(x) for x in self.root.glob('**')]

    def __eq__(self, other):
        if not isinstance(other, Registry):
            return False
        if other.name != self.name:
            return False
        if other.root != self.root:
            return False
        return True


class InMemoryRegistry(Registry):
    def __init__(self, name, priority: int = 100, **components):
        self.name = name
        self.root = '/proc/memory/'  # Something pathlike is needed?
        self.version = mccode_antlr_version()
        self.components = {k: v for k, v in components.items()}
        self.priority = priority

    def add(self, name: str, definition: str):
        self.components[name] = definition

    def add_comp(self, name: str, definition: str):
        if not name.lower().endswith('.comp'):
            name += '.comp'
        self.add(name, definition)

    def add_instr(self, name: str, definition: str):
        if not name.lower().endswith('.instr'):
            name += '.instr'
        self.add(name, definition)

    def filenames(self) -> list[str]:
        return list(self.components.keys())

    def fullname(self, name: str, ext: str | None = None):
        full_name = name if ext is None else name + ext
        return full_name if full_name in self.components else None

    def known(self, name: str, ext: str | None = None):
        full_name = self.fullname(name, ext=ext)
        if full_name is not None and full_name in self.components:
            return True
        return False

    def contents(self, name: str, ext: str | None = None):
        full_name = self.fullname(name, ext=ext)
        if full_name is not None and full_name in self.components:
            return self.components[full_name]
        raise KeyError(f'InMemoryRegistry does not know of {name if ext is None else name + ext}')


def ordered_registries(registries: list[Registry]):
    """Sort the registries by their priority"""
    return sorted(registries, key=lambda x: x.priority, reverse=True)


def registries_match(registry: Registry, spec):
    if isinstance(spec, Registry):
        return registry == spec
    parts = spec.split()
    path = Path(parts[1] if len(parts) == 2 else parts[0])
    if path.exists() and registry.root == path.resolve():
        return True
    url = parts[1] if len(parts) >= 2 else parts[0]
    if registry.pooch is not None and registry.pooch.base_url == url:
        return True
    return False


def registry_from_specification(spec: str):
    """Construct a Local or Remote Registry instance from a specification string

    Expected specifications are:

    1. {resolvable folder path}
    2. {name} {resolvable folder path}
    3. {name} {resolvable url} {resolvable file path}
    4. {name} {resolvable url} {version} {registry file name}

    The first two variants make a LocalRegistry, which searches the provided directory for files.
    The third makes a ModuleRemoteRegistry using pooch. The resolvable file path should point at a Pooch registry file.
    The fourth makes a GitHubRegistry, which uses the specific folder structure of GitHub
    """
    if isinstance(spec, Registry):
        return spec
    parts = spec.split()
    if len(parts) == 0:
        return None
    elif len(parts) == 1:
        p1, p2, p3, p4 = parts[0], parts[0], None, None
    elif len(parts) < 4:
        p1, p2, p3, p4 = parts[0], parts[1], None if len(parts) < 3 else parts[2], None
    else:
        p1, p2, p3, p4 = parts[0], parts[1], parts[2], parts[3]
    print(f"Constructing registry from {p1=} {p2=} {p3=} {p4=}  [{type(p1)=} {type(p2)=} {type(p3)=} {type(p4)=}]")
    # convert string literals to strings:
    p1 = p1[1:-1] if p1.startswith('"') and p1.endswith('"') else p1
    p2 = p2[1:-1] if p2.startswith('"') and p2.endswith('"') else p2
    p3 = p3[1:-1] if p3 is not None and p3.startswith('"') and p3.endswith('"') else p3
    p4 = p4[1:-1] if p4 is not None and p4.startswith('"') and p4.endswith('"') else p4

    if Path(p2).exists() and Path(p2).is_dir():
        return LocalRegistry(p1, str(Path(p2).resolve()))

    # (simple) URL validation:
    if not simple_url_validator(p2, file_ok=True):
        return None

    if Path(p3).exists() and Path(p3).is_file():
        return ModuleRemoteRegistry(p1, p2, Path(p3).resolve().as_posix())

    if p4 is not None:
        return GitHubRegistry(p1, p2, p3, p4)

    return None


def _mccode_pooch_registries():
    def get_remote_repository_version_tags(url):
        import re
        import git
        g = git.cmd.Git()
        res = g.ls_remote(url, sort='-v:refname', tags=True)
        ex = re.compile(r'v\d+(?:\.\d+(?:\.\d+)?)?')
        return ex.findall(res)

    def source_registry_tag():
        from mccode_antlr.config import config
        requested_tag = config['mccode_pooch']['tag'].as_str_expanded()
        registry_url = config['mccode_pooch']['registry'].as_str_expanded()
        source_url = config['mccode_pooch']['source'].as_str_expanded()

        known_tags = get_remote_repository_version_tags(registry_url)
        if requested_tag.lower() == 'latest':
            requested_tag = known_tags[0]
        elif requested_tag not in known_tags:
            raise RuntimeError(f"The specified version tag, {requested_tag}, is not available in {registry_url}")
        return source_url, registry_url, requested_tag

    src, reg, tag = source_registry_tag()

    mc, mx, lib = [GitHubRegistry(name, src, tag, registry=reg, priority=-10) for name in ('mcstas', 'mcxtrace', 'libc')]
    return mc, mx, lib

MCSTAS_REGISTRY, MCXTRACE_REGISTRY, LIBC_REGISTRY = _mccode_pooch_registries()



from textwrap import dedent
from .c_listener import CDeclarator
from ..instr import Instr


def header_pre_runtime(
        is_mcstas: bool,
        source: Instr,
        runtime: dict,
        config: dict,
        uservars: list[CDeclarator]
):
    from datetime import datetime
    from mccode_antlr import version

    def jump_line(instance, jump):
        return f'long Jump_{instance.name}_{jump.target}; /* the JUMP connection <from>_<to> */'

    jump_string = '\n'.join([jump_line(i, j) for i in source.components for j in i.jump if j.iterate])

    if is_mcstas:
        particle_struct = '\n'.join([
            "  double vx,vy,vz; /* velocity [m/s] */",
            "  double sx,sy,sz; /* spin [0-1] */",
            "  int mcgravitation; /* gravity-state */",
            "  void *mcMagnet;    /* precession-state */",
            "  int allow_backprop; /* allow backprop */"])
    else:
        particle_struct = '\n'.join([
            "  double kx,ky,kz; /* wave-vector */",
            "  double phi, Ex,Ey,Ez; /* phase and electrical field */"])

    # Append variables from instr USERVARS block to particle struct
    # Also store these strings in the appropriate instrument list for later def/undef as state variables

    uservar_string = '//user variables and comp - injections:\n' if len(uservars) else ''
    uservar_string += '\n'.join([f'  {x};' for x in uservars])

    # Shouldn't we exclude array-valued names here?
    getvar = '\n'.join(
        [f'  if(!str_comp("{x.name}",name)){{rval=*((double*)(&(p->{x.name})));s=0;}}' for x in uservars])

    # Array valued names seem OK here, since a user can type-cast correctly
    getvar_void = '\n'.join(
        [f'  if (!str_comp("{x.name}", name)){{rval=(void * ) & (p->{x.name});s=0;}}' for x in uservars])

    # For non-array values we can safely use memcpy (probably)
    setvar_void = '\n'.join(
        [f'  if(!str_comp("{x.name}",name)){{memcpy(&(p->{x.name}), value, sizeof({x.dtype})); rval=0;}}'
         for x in uservars if not x.is_array and not x.is_pointer])

    # an array -- need to know how many elements we're copying
    setvar_void_array = '\n'.join(
        [f'  if(!str_comp("{x.name}",name)){{memcpy(&(p->{x.name}), value, elements * sizeof({x.dtype})); rval=0;}}'
         for x in uservars if x.is_array or x.is_pointer])

    getuservar_byid = '\n'.join(
        [f'  case {i}: {{rval=*( (double *)(&(p->{x.name})) ); s=0; break;}}' for i, x in enumerate(uservars)])

    uservar_init = ''
    for x in uservars:
        if (x.dtype not in ('double', 'MCNUM', 'int') and x.init is None) or x.is_pointer or x.is_array:
            array_str = ' array' if x.is_pointer or x.is_array else ''
            print(f'\nWARNING:\n --> USERVAR {x.name} is of type {x.dtype}{array_str}')
            print('  --> and may need specific per-particle initialization through an EXTEND block!\n')
        else:
            uservar_init += f'\np->{x.name}={0 if x.init is None else x.init};'

    contents = dedent(f"""/* Automatically generated file. Do not edit.
     * Format:     ANSI C source code
     * Creator:    {runtime.get("fancy")} <{runtime.get("url")}>
     * Generator:  mccode-antlr {version()} <https://github.com/McStasMcXtrace/mccode-antlr.git>
     * Instrument: {source.source} ({source.name})
     * Date: {datetime.now()}
     * File: {config.get('output')}
     * CFLAGS={' '.join(set(source.flags))}
     */
     
    /* In case of cl.exe on Windows, suppress warnings about #pragma acc
       Transferred from https://github.com/McStasMcXtrace/McCode/commit/0e2785a2d3fd742d46597139234dbc47e56344bb 
    */
    #ifdef _MSC_EXTENSIONS
    #pragma warning(disable: 4068)
    #endif
     
    #define MCCODE_STRING "{runtime.get("fancy")}"
    #define FLAVOR "{runtime.get("name", "none")}"
    #define FLAVOR_UPPER "{runtime.get("name", "none").upper()}"
    {'#define MC_USE_DEFAULT_MAIN' if config.get('default_main') else ''}
    {'#define MC_TRACE_ENABLED' if config.get('enable_trace') else ''}
    {'#define MC_PORTABLE' if config.get('portable') else ''}
    
    #include <string.h>
    
    typedef double MCNUM;
    typedef struct {{MCNUM x, y, z;}} Coords;
    typedef MCNUM Rotation[3][3];
    #define MCCODE_BASE_TYPES
    
    #ifndef MC_NUSERVAR
    #define MC_NUSERVAR 10
    #endif
    
    /* Particle JUMP control logic */
    struct particle_logic_struct {{
      int dummy;{jump_string}
    }};
    struct _struct_particle {{
      double x,y,z; /* position [m] */
    {particle_struct}
      /* Generic Temporaries: */
      /* May be used internally by components e.g. for special */
      /* return-values from functions used in trace, thusreturned via */
      /* particle struct. (Example: Wolter Conics from McStas, silicon slabs.) */
      double _mctmp_a; /* temp a */
      double _mctmp_b; /* temp b */
      double _mctmp_c; /* temp c */
      unsigned long randstate[7];
      double t, p;    /* time, event weight */
      long long _uid;  /* Unique event ID */
      long _index;     /* component index where to send this event */
      long _absorbed;  /* flag set to TRUE when this event is to be removed/ignored */
      long _scattered; /* flag set to TRUE when this event has interacted with the last component instance */
      long _restore;   /* set to true if neutron event must be restored */
      long flag_nocoordschange;   /* set to true if particle is jumping */
      struct particle_logic_struct _logic;
      {uservar_string}
    }};
    typedef struct _struct_particle _class_particle;
    
    _class_particle _particle_global_randnbuse_var;
    _class_particle* _particle = &_particle_global_randnbuse_var;
    
    // Below lines relating to mcgenstate / setstate are in principle McStas - centric, we ought to generate
    //this function based on "project"
    #pragma acc routine
    _class_particle mcgenstate(void);
    #pragma acc routine
    _class_particle mcsetstate(double x, double y, double z, double vx, double vy, double vz,
        double t, double sx, double sy, double sz, double p, int mcgravitation, void *mcMagnet, int mcallowbackprop);
    
    extern int mcgravitation;      /* flag to enable gravitation */
    #pragma acc declare create ( mcgravitation )
    int mcallowbackprop;        
    #pragma acc declare create ( mcallowbackprop )
    
    _class_particle mcgenstate(void) {{
      _class_particle particle = mcsetstate(0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, mcgravitation, NULL, mcallowbackprop);
      return(particle);
    }}
    /*Generated user variable handlers:*/
    
    #pragma acc routine
    double particle_getvar(_class_particle *p, char *name, int *suc);
    
    #ifdef OPENACC
    #pragma acc routine
    int str_comp(char *str1, char *str2);
    #endif
    
    double particle_getvar(_class_particle *p, char *name, int *suc){{
    #ifndef OPENACC
    #define str_comp strcmp
    #endif
      int s=1;
      double rval=0;
      if(!str_comp("x",name)){{rval=p->x;s=0;}}
      if(!str_comp("y",name)){{rval=p->y;s=0;}}
      if(!str_comp("z",name)){{rval=p->z;s=0;}}
      if(!str_comp("vx",name)){{rval=p->vx;s=0;}}
      if(!str_comp("vy",name)){{rval=p->vy;s=0;}}
      if(!str_comp("vz",name)){{rval=p->vz;s=0;}}
      if(!str_comp("sx",name)){{rval=p->sx;s=0;}}
      if(!str_comp("sy",name)){{rval=p->sy;s=0;}}
      if(!str_comp("sz",name)){{rval=p->sz;s=0;}}
      if(!str_comp("t",name)){{rval=p->t;s=0;}}
      if(!str_comp("p",name)){{rval=p->p;s=0;}}
      if(!str_comp("_mctmp_a",name)){{rval=p->_mctmp_a;s=0;}}
      if(!str_comp("_mctmp_b",name)){{rval=p->_mctmp_b;s=0;}}
      if(!str_comp("_mctmp_c",name)){{rval=p->_mctmp_c;s=0;}}
    {getvar}  
      if (suc!=0x0) {{*suc=s;}}
      return rval;
    }}
    
    #pragma acc routine
    void* particle_getvar_void(_class_particle *p, char *name, int *suc);\n
    #ifdef OPENACC
    #pragma acc routine
    int str_comp(char *str1, char *str2);
    #endif
    
    void* particle_getvar_void(_class_particle *p, char *name, int *suc){{
    #ifndef OPENACC
    #define str_comp strcmp
    #endif
      int s=1;
      void* rval=0;
      if(!str_comp("x",name)) {{rval=(void*)&(p->x); s=0;}}
      if(!str_comp("y",name)) {{rval=(void*)&(p->y); s=0;}}
      if(!str_comp("z",name)) {{rval=(void*)&(p->z); s=0;}}
      if(!str_comp("vx",name)){{rval=(void*)&(p->vx);s=0;}}
      if(!str_comp("vy",name)){{rval=(void*)&(p->vy);s=0;}}
      if(!str_comp("vz",name)){{rval=(void*)&(p->vz);s=0;}}
      if(!str_comp("sx",name)){{rval=(void*)&(p->sx);s=0;}}
      if(!str_comp("sy",name)){{rval=(void*)&(p->sy);s=0;}}
      if(!str_comp("sz",name)){{rval=(void*)&(p->sz);s=0;}}
      if(!str_comp("t",name)) {{rval=(void*)&(p->t); s=0;}}
      if(!str_comp("p",name)) {{rval=(void*)&(p->p); s=0;}}
    {getvar_void}
      if (suc!=0x0) {{*suc=s;}}
      return rval;
    }}
    
    #pragma acc routine
    int particle_setvar_void(_class_particle *, char *, void*);\n
    int particle_setvar_void(_class_particle *p, char *name, void* value){{
    #ifndef OPENACC
    #define str_comp strcmp
    #endif
      int rval=1;
      if(!str_comp("x",name)) {{memcpy(&(p->x),  value, sizeof(double)); rval=0;}}
      if(!str_comp("y",name)) {{memcpy(&(p->y),  value, sizeof(double)); rval=0;}}
      if(!str_comp("z",name)) {{memcpy(&(p->z),  value, sizeof(double)); rval=0;}}
      if(!str_comp("vx",name)){{memcpy(&(p->vx), value, sizeof(double)); rval=0;}}
      if(!str_comp("vy",name)){{memcpy(&(p->vy), value, sizeof(double)); rval=0;}}
      if(!str_comp("vz",name)){{memcpy(&(p->vz), value, sizeof(double)); rval=0;}}
      if(!str_comp("sx",name)){{memcpy(&(p->sx), value, sizeof(double)); rval=0;}}
      if(!str_comp("sy",name)){{memcpy(&(p->sy), value, sizeof(double)); rval=0;}}
      if(!str_comp("sz",name)){{memcpy(&(p->sz), value, sizeof(double)); rval=0;}}
      if(!str_comp("p",name)) {{memcpy(&(p->p),  value, sizeof(double)); rval=0;}}
      if(!str_comp("t",name)) {{memcpy(&(p->t),  value, sizeof(double)); rval=0;}}
    {setvar_void}  
      return rval;
    }}
    
    #pragma acc routine
    int particle_setvar_void_array(_class_particle *, char *, void*, int);
    
    int particle_setvar_void_array(_class_particle *p, char *name, void* value, int elements){{
    #ifndef OPENACC
    #define str_comp strcmp
    #endif
      int rval=1;
    {setvar_void_array}
      return rval;
    }}
    
    // Function to handle a particle restore of physical particle params
    #pragma acc routine
    void particle_restore(_class_particle *p, _class_particle *p0);
    void particle_restore(_class_particle *p, _class_particle *p0) {{
      p->x  = p0->x;  p->y  = p0->y;  p->z  = p0->z;
      p->vx = p0->vx; p->vy = p0->vy; p->vz = p0->vz;
      p->sx = p0->sx; p->sy = p0->sy; p->sz = p0->sz;
      p->t = p0->t;  p->p  = p0->p;
      p->_absorbed=0; p->_restore=0;
    }}
    
    #pragma acc routine
    double particle_getuservar_byid(_class_particle *p, int id, int *suc){{
      int s=1;
      double rval=0;
      switch(id){{
    {getuservar_byid}
      }}
      if (suc!=0x0) {{*suc=s;}}
      return rval;
    }}
    
    #pragma acc routine
    void particle_uservar_init(_class_particle *p){{
    {uservar_init}
    }}
    """)
    return contents


def header_post_runtime(source, runtime: dict, config: dict, include_path):
    from ..common.utilities import escape_str_for_c

    def source_file_contents():
        from pathlib import Path
        from os import access, R_OK
        if source.source is None and config.get('embed_instrument_file'):
            raise RuntimeError('Requested source embedding, but no source file provided!')
        elif source.source is None:
            return 'Instrument source code is not embedded in this executable.'
        path = Path(source.source)
        if config.get('embed_instrument_file') and path.exists() and access(path, R_OK):
            with path.open('r') as file:
                return escape_str_for_c(file.read())
        message = f"Instrument {source.name} source code "
        message += f"from {escape_str_for_c(source.source)} is not embedded in this executable.\\n"
        message += f"Use --source option when running {config.get('flavor')}"
        return message

    main_file_string = 'int main(int argc, char *argv[]){return mccode_main(argc, argv);}'
    contents = dedent(f"""
    /* *****************************************************************************
     * Start of instrument '{source.name}' generated code
    ***************************************************************************** */
    
    #ifdef MC_TRACE_ENABLED
    int traceenabled = 1;
    #else
    int traceenabled = 0;
    #endif
    #define {runtime.get("name", "none").upper()} "{escape_str_for_c(str(include_path))}"
    int   defaultmain         = {1 if config.get("default_main") else 0};
    char  instrument_name[]   = "{source.name}";
    char  instrument_source[] = "{escape_str_for_c(source.source)}";
    char *instrument_exe      = NULL; /* will be set to argv[0] in main */
    char instrument_code[] = "{source_file_contents()}";
    {main_file_string if config.get('default_main') else ''}
    """)
    return contents



def cogen_getvarpars_fct(instr):
    lines = [
        "void* _getvar_parameters(char* compname)",
        "/* enables settings parameters based use of the GETPAR macro */",
        "{",
        "  #ifdef OPENACC",
        "    #define strcmp(a,b) str_comp(a,b)",
        "  #endif",
    ]
    for comp in instr.components:
        lines.append(f'  if (!strcmp(compname, "{comp.name}")) return (void *) &(_{comp.name}_var._parameters);')

    lines.extend([
        '  return 0;',
        '}',
        ''
    ])
    return '\n'.join(lines)


def cogen_getparticlevar_fct(uservars):
    lines = [
        "void* _get_particle_var(char *token, _class_particle *p)",
        "/* enables setpars based use of GET_PARTICLE_DVAR macro and similar */",
        "{",
    ]
    for var in uservars:
        lines.append(f'  if (!strcmp(token, "{var.name}")) return (void *) &(p->{var.name});')

    lines.extend([
        "  return 0;",
        "}",
        "",
    ])
    return '\n'.join(lines)


def cogen_getcompindex_fct(instr):
    lines = [
        "int _getcomp_index(char* compname)",
        "/* Enables retrieving the component position & rotation when the index is not known.",
        " * Component indexing into MACROS, e.g., POS_A_COMP_INDEX, are 1-based! */",
        "{",
    ]
    for index, comp in enumerate(instr.components):
        lines.append(f'  if (!strcmp(compname, "{comp.name}")) return {1+index};')

    lines.extend([
        "  return -1;",
        "}",
        "",
    ])
    return '\n'.join(lines)



from __future__ import annotations
from loguru import logger
from enum import Enum
from pathlib import Path


class Target(Enum):
    python = 0
    cpp = 1
    def __str__(self):
        if self == Target.python:
            return 'Python3'
        elif self == Target.cpp:
            return 'Cpp'
        raise ValueError(f'Unknown target {self}')


class Feature(Enum):
    listener = 0
    visitor = 1


def rebuild_language(grammar_file,
                     target: Target,
                     features: list[Feature],
                     verbose=False,
                     output=None
                     ):
    from pathlib import Path
    from subprocess import Popen, PIPE
    from antlr4_tool_runner import initialize_paths, install_jre_and_antlr
    from antlr4_tool_runner import process_args

    def antlr4_version():
        return process_args()[1]

    if not isinstance(grammar_file, Path):
        grammar_file = Path(grammar_file)

    args =[
        f'-Dlanguage={target}',
        '-visitor' if Feature.visitor in features else '-no-visitor',
        '-listener' if Feature.listener in features else '-no-listener',
        '-o', output or str(grammar_file.parent / str(target)),
        str(grammar_file)
    ]

    if verbose:
        logger.info(f'Building language files for {grammar_file}')

    # The following copies the implementation of antlr4_tool_runner.tool, which pulls `args` from the system argv list
    # Setup:
    initialize_paths()
    version = antlr4_version()
    jar, java = install_jre_and_antlr(version)
    # Call antlr4
    p = Popen([java, '-cp', jar, 'org.antlr.v4.Tool'] + args, stdout=PIPE, stderr=PIPE)
    out, err = [x.decode('UTF-8') for x in p.communicate()]
    if err:
        print(err, end='')
    if out:
        print(out, end='')


def language_present_and_up_to_date(grammar_file, newest, features, path, verbose=False):
    from pathlib import Path
    if not isinstance(grammar_file, Path):
        grammar_file = Path(grammar_file)

    if not grammar_file.exists():
        raise RuntimeError(f"The grammar file {grammar_file} does not exist.")

    lexer_file = path.joinpath(f'{grammar_file.stem}Lexer.py')
    parser_file = path.joinpath(f'{grammar_file.stem}Parser.py')
    listener_file = path.joinpath(f'{grammar_file.stem}Listener.py')
    visitor_file = path.joinpath(f'{grammar_file.stem}Visitor.py')

    generated_files = (lexer_file, parser_file)
    if Feature.visitor in features:
        generated_files += (visitor_file, )
    if Feature.listener in features:
        generated_files += (listener_file, )

    if not all(x.exists() for x in generated_files):
        if verbose:
            logger.info(f'Not all language files exist for {grammar_file}')
        return False

    if any(x.stat().st_mtime < newest for x in generated_files):
        if verbose:
            logger.info(f'Not all language files up-to-date for {grammar_file}')
        return False

    if verbose:
        logger.info(f'Language files for {grammar_file} are up-to-date')
    return True


# def rebuild_speedy_language(grammar_file, features: list[Feature], output: Path, verbose=False):
#     from speedy_antlr_tool import generate
#
#     cpp_output_dir = grammar_file.parent / str(Target.cpp)
#
#     rebuild_language(grammar_file, Target.cpp, features, verbose=verbose, output=cpp_output_dir)
#     rebuild_language(grammar_file, Target.python, features, verbose=verbose, output=output)
#
#     py_parser_path = output / f'{grammar_file.stem}Parser.py'
#     generate(str(py_parser_path), cpp_output_dir)


def ensure_language_up_to_date(
        grammar: str,
        *,
        target: Target,
        features: list[Feature],
        deps=None,
        verbose=False,
):
    """Ensure the ANTLR parsed language files are up-to-date."""
    from pathlib import Path
    # This file is at src/grammar/builder.py, it's parent is src/grammar
    # which the grammar defining files are under
    grammar_file = Path(__file__).parent / f'{grammar}.g4'
    # and we want to put Python files under src/mccode_antlr/grammar
    output_path = Path(__file__).parent.parent / "mccode_antlr" / "grammar"

    newest = grammar_file.stat().st_mtime
    if deps:
        for dep in deps:
            newest = max(newest, Path(__file__).parent.joinpath(f'{dep}.g4').stat().st_mtime)

    if not language_present_and_up_to_date(grammar_file, newest, features, output_path, verbose=verbose):
        rebuild_language(grammar_file, target, features, output=output_path, verbose=verbose)


def main():
    from argparse import ArgumentParser

    parser = ArgumentParser(prog="mccode-antlr-build", description='Ensure ANTLR files are up-to-date')
    parser.add_argument('-v ', '--verbose', action='store_true', help='Print out more information')
    args = parser.parse_args()
    verbose = args.verbose

    ensure_language_up_to_date('McComp', target=Target.python, features=[Feature.visitor], deps=('McCommon', 'c99'), verbose=verbose)
    ensure_language_up_to_date('McInstr', target=Target.python, features=[Feature.visitor], deps=('McCommon', 'c99'), verbose=verbose)
    ensure_language_up_to_date('C', target=Target.python, features=[Feature.visitor, Feature.listener], verbose=verbose)

if __name__ == '__main__':
    main()



from unittest import TestCase
from .utils import parse_instr_string


class TestInstrCopy(TestCase):
    def test_copy(self):
        from mccode_antlr.instr import Instr
        from mccode_antlr.common import InstrumentParameter, Expr, Value, DataType
        instr_source = """
        DEFINE INSTRUMENT test_copy(par0=3.14159, double par1 = 49, int par2 =     1010110
            , string par3="this is a long string with spaces", 

            string par4, int par5, double par6, par7)
        TRACE
        COMPONENT first = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT second = Arm() AT (0, 0, 1) RELATIVE first ROTATED (0, 90, 0) RELATIVE first
        END
        """
        instr = parse_instr_string(instr_source)
        instr_copy = instr.copy()
        self.assertTrue(isinstance(instr_copy, Instr))
        self.assertEqual(instr.name, instr_copy.name)
        self.assertEqual(len(instr.parameters), len(instr_copy.parameters))
        self.assertEqual(len(instr.components), len(instr_copy.components))
        for p, p_copy in zip(instr.parameters, instr_copy.parameters):
            self.assertTrue(isinstance(p_copy, InstrumentParameter))
            self.assertEqual(p.name, p_copy.name)
            self.assertEqual(p.value, p_copy.value)

        instr_copy.add_parameter(InstrumentParameter('par8', '', Expr.float(1.6189)))
        self.assertEqual(len(instr.parameters), len(instr_copy.parameters) - 1)



def _import_component_language():
    # from .sa_mccomp import parse, SA_ErrorListener
    from .mccomp_parse import parse, ErrorListener
    from .McCompVisitor import McCompVisitor
    from .McCompParser import McCompParser
    return parse, ErrorListener, McCompVisitor, McCompParser


def _import_instrument_language():
    # from .sa_mcinstr import parse, SA_ErrorListener
    from .mcinstr_parse import parse, ErrorListener
    from .McInstrVisitor import McInstrVisitor
    from .McInstrParser import McInstrParser
    return parse, ErrorListener, McInstrVisitor, McInstrParser


def _import_c_language():
    from .CLexer import CLexer
    from .CParser import CParser
    from .CListener import CListener
    from .CVisitor import CVisitor
    return CLexer, CParser, CListener, CVisitor


# Import the classes defined in the language files
McComp_parse, McComp_ErrorListener, McCompVisitor, McCompParser = _import_component_language()
McInstr_parse, McInstr_ErrorListener, McInstrVisitor, McInstrParser = _import_instrument_language()
CLexer, CParser, CListener, CVisitor = _import_c_language()

# And set only their names to be exported:
__all__ = [
    'McComp_parse',
    'McComp_ErrorListener',
    'McCompParser',
    'McCompVisitor',
    'McInstr_parse',
    'McInstr_ErrorListener',
    'McInstrParser',
    'McInstrVisitor',
    'CLexer',
    'CParser',
    'CListener',
    'CVisitor',
]



from __future__ import annotations
from functools import cache
from unittest import TestCase


@cache
def check_for_mccode_antlr_compiler(which: str) -> bool:
    import subprocess
    from loguru import logger
    from mccode_antlr.config import config
    cc = config
    for key in which.split('/'):
        cc = cc[key]
    cc = cc.get(str)

    try:
        subprocess.run([cc, '--version'], check=True)
        return True
    except FileNotFoundError:
        logger.info(f'Expected compiler {cc} not found.')
        logger.info('Provide alternate C compiler via MCCODE_ANTLR_CC environment variable')
    return False


@cache
def simple_instr_compiles(which: str) -> bool:
    if check_for_mccode_antlr_compiler(which):
        try:
            from mccode_antlr.loader import parse_mcstas_instr
            instr = parse_mcstas_instr("define instrument check() trace component a = Arm() at (0,0,0) absolute end")
            compile_and_run(instr, "-n 1", run=False, target={'acc': which == 'acc'})
        except RuntimeError:
            return False
        except FileNotFoundError:
            return False
    return True


def compiled(method, compiler: str | None = None):
    if compiler is None:
        # Basic compiled instruments only need the 'cc' compiler specified in the config file
        compiler = 'cc'

    def wrapper(*args, **kwargs):
        if simple_instr_compiles(compiler):
            method(*args, **kwargs)
        elif isinstance(args[0], TestCase):
            args[0].skipTest(f'Skipping due to lack of working ${compiler}')

    return wrapper


def gpu_only(method):
    # GPU compiled instruments need the specific OpenACC compiler
    # **PLUS** they need to _actually_ have the openACC header (macOS and Windows don't use different compilers)
    return compiled(method, 'acc')


def mpi_only(method):
    # MPI compiled instruments need the specified compiler
    return compiled(method, 'mpi/cc')


@cache
def mcpl_config_available():
    try:
        import subprocess
        subprocess.run(['mcpl-config', '--version'], check=True)
    except FileNotFoundError:
        return False
    return True


def mcpl_compiled(method):
    def wrapper(*args, **kwargs):
        if mcpl_config_available() and simple_instr_compiles('cc'):
            method(*args, **kwargs)
        elif isinstance(args[0], TestCase):
            msg = "working compiler" if mcpl_config_available() else "mcpl interface"
            args[0].skipTest(f'Skipping due to missing {msg}')

    return wrapper


def compile_and_run(instr, parameters, run=True, dump_source=True, target: dict | None = None, config: dict | None = None):
    from pathlib import Path
    from tempfile import TemporaryDirectory
    from mccode_antlr.translators.target import MCSTAS_GENERATOR
    from mccode_antlr.run import mccode_compile, mccode_run_compiled

    kwargs = dict(generator=MCSTAS_GENERATOR, target=target, config=config, dump_source=dump_source)

    with TemporaryDirectory() as directory:
        binary, target = mccode_compile(instr, directory, **kwargs)
        # The runtime output directory used *can not* exist for McStas/McXtrace to work properly.
        # So find a name inside this directory that doesn't exist (any name should work)
        return mccode_run_compiled(binary, target, Path(directory).joinpath('t'), parameters) if run else (None, None)



from unittest import TestCase
from loguru import logger
from .compiled import compiled, mcpl_compiled, compile_and_run


class TestCompiledInstr(TestCase):
    @compiled
    def test_one_axis(self):
        from math import pi, asin, sqrt
        from mccode_antlr.loader import parse_mcstas_instr
        d_spacing = 3.355  # (002) for Highly-ordered Pyrolytic Graphite
        mean_energy = 5.0
        energy_width = 0.5
        mean_ki = sqrt(mean_energy / 2.0722)
        instr = f"""
        DEFINE INSTRUMENT splitRunTest(a1=0, a2=0, virtual_source_x=0.05, virtual_source_y=0.1, string newname)
        TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT source = Source_simple(yheight=2*virtual_source_y, xwidth=0.2, dist=1.5, focus_xw=0.06, focus_yh=0.12,
                                         E0={mean_energy}, dE={energy_width})
                           AT (0, 0, 0) RELATIVE origin
        COMPONENT m0 = PSD_monitor(xwidth=0.1, yheight=0.15, nx=100, ny=160, restore_neutron=1) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT guide = Guide_gravity(w1 = 0.06, h1 = 0.12, w2 = 0.05, h2 = 0.1, l = 30, m = 4) 
                          AT (0, 0, 1.5) RELATIVE  PREVIOUS
        COMPONENT guide_end = Arm() AT (0, 0, 30) RELATIVE PREVIOUS
        COMPONENT m1 = PSD_monitor(xwidth=0.1, yheight=0.15, nx=100, ny=160, restore_neutron=1) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT aperture = Slit(xwidth=virtual_source_x, yheight=virtual_source_y) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT split_at = Arm() AT (0, 0, 0.0001) RELATIVE PREVIOUS
        COMPONENT m3 = PSD_monitor(xwidth=0.1, yheight=0.15, nx=100, ny=160, restore_neutron=1) AT (0, 0, 0.01) RELATIVE PREVIOUS
        COMPONENT mono_point = Arm() AT (0, 0, 0.8) RELATIVE split_at
        METADATA "txt" "something" %{{
            This is some unparsed metadata that will be included as a literal string in the instrument.
        %}}
        COMPONENT mono = Monochromator_curved(zwidth = 0.02, yheight = 0.02, NH = 13, NV = 7, DM={d_spacing}) 
                         AT (0, 0, 0) RELATIVE  mono_point ROTATED (0, a1, 0) RELATIVE mono_point
        COMPONENT sample_arm = Arm() AT (0, 0, 0) RELATIVE mono_point ROTATED (0, a2, 0) RELATIVE mono_point
        COMPONENT detector = Monitor(xwidth=0.1, yheight=0.15, restore_neutron=1) AT (0, 0, 0.8) RELATIVE sample_arm
        COMPONENT lmon = L_monitor(filename=newname) AT (0, 0, 0.001) RELATIVE PREVIOUS
        END
        """
        instr = parse_mcstas_instr(instr)

        a1 = asin(pi / d_spacing / mean_ki) * 180 / pi
        self.assertAlmostEqual(a1, 37.0722, 4)
        parameters = f'a1={a1} a2={2 * a1} -n 1000000'

        std_output, dats = compile_and_run(instr, parameters)
        self.assertEqual(len(dats), 5)
        self.assertEqual(dats['m0'].data.shape, (3, 160, 100))
        self.assertEqual(dats['m1'].data.shape, (3, 160, 100))
        self.assertEqual(dats['m3'].data.shape, (3, 160, 100))
        self.assertEqual(dats['detector'].data.shape, (3, ))

        # Moving farther from the source means less (but finite) intensity in equivalent monitors
        self.assertTrue(sum(sum(dats['m0']['I'])) > sum(sum(dats['m1']['I'])) > sum(sum(dats['m3']['I'])) > 0)
        # The detector has been positioned correctly to collect intensity
        self.assertTrue(dats['detector']['I'] > 0)

    @compiled
    def test_assembled_parameters(self):
        """Check that setting an instance parameter to a value that is an instrument parameter name works"""
        from mccode_antlr.assembler import Assembler
        from mccode_antlr.reader import MCSTAS_REGISTRY
        assembler = Assembler('assembled_parameters_test_instr', registries=[MCSTAS_REGISTRY])
        assembler.parameter("double par0 = 3.14159")
        origin = assembler.component("origin", "Progress_bar", at=[0, 0, 0])
        left = assembler.component('left', 'Slit', at=([0, 0, 1], origin), rotate=[0, 90, 0],
                                   parameters=dict(xwidth='par0', yheight='2*fmod(par0, 0.1)'))

        compile_and_run(assembler.instrument, None, run=False)

    @compiled
    def test_vector_component_parameter(self):
        """Some components can use vector parameters, which must be initialized by initializer lists"""
        from mccode_antlr.common import Value
        from mccode_antlr.loader import parse_mcstas_instr
        # TODO Update the registry with the new Conics_* components, then update this test (if necessary)
        instr = """
        DEFINE INSTRUMENT test_vector_component_parameter() TRACE
        COMPONENT cEH = Conics_EH(R0=0.99, alpha=6.07, W=0.003, m=3, nshells=4, focal_length_u=10, focal_length_d=10,
                                  radii={0.05236, 0.03, 0.01, 0.0031416}, le=0.25, lh=0.25, disk=1)
        AT (0,0,0) ABSOLUTE
        COMPONENT cEH2 = Conics_EH(R0=0.99, alpha=6.07, W=0.003, m=3, nshells=0, focal_length_u=10, focal_length_d=10,
                                   le=0.25, lh=0.25, disk=0)
        AT (0,0,1) ABSOLUTE
        END
        """
        instr = parse_mcstas_instr(instr)
        self.assertEqual(instr.components[0].get_parameter('radii').value, Value([0.05236, 0.03, 0.01, 0.0031416]))

        try:
            compile_and_run(instr, '-n 1000', run=True)
        except RuntimeError as e:
            logger.error(f'Failed to compile instrument: {e}')
            self.fail(f'Failed to compile instrument {e}')

    @compiled
    def test_split_broken_reference_compiles(self):
        from mccode_antlr.loader import parse_mcstas_instr
        from textwrap import dedent
        instr = parse_mcstas_instr(dedent("""\
        DEFINE INSTRUMENT test_tof(phase/"degree"=0, ang/"degree"=0)
        TRACE
        COMPONENT Origin = Progress_bar() AT (0, 0, 0) ABSOLUTE 
        COMPONENT ESS_source = ESS_butterfly(
            sector="W", beamline=4, yheight=0.03, cold_frac=0.5, dist=1, focus_xw=0.01, focus_yh=0.01, 
            c_performance=1.0, t_performance=1.0, Lmin=1, Lmax=2, tmax_multiplier=1.5, n_pulses=1, acc_power=2.0
        ) AT (0, 0, 0) ABSOLUTE 
        COMPONENT guide = Guide_gravity(w1=0.01, w2=0.05, h1=0.01, h2=0.05, l=8.0, m=3.5, G=-9.82) AT (0, 0, 1) ABSOLUTE 
        COMPONENT guide_end = Arm() AT (0, 0, 8.0) RELATIVE guide
        COMPONENT chopper = DiskChopper(radius=0.35, nu=14, phase=phase, theta_0=115) AT (0, 0, 0.01) RELATIVE guide_end
        COMPONENT rotate = Arm() AT (0, 0, 0) RELATIVE chopper ROTATED (0, ang, 0) RELATIVE chopper
        COMPONENT aperture = Slit(xwidth=0.05, yheight=0.05) AT (0, 0, 1) RELATIVE rotate
        COMPONENT split_at = Arm() AT (0, 0, 1e-08) RELATIVE rotate
        COMPONENT sample = Incoherent(
            radius=0.005, yheight=0.02, thickness=0.001, focus_ah=2.0, focus_aw=2.0, target_x=1.0, target_y=0.0, 
            target_z=0.0, Etrans=0.0, deltaE=0.2
        ) AT (0, 0, 1) RELATIVE aperture
        END"""))
        _, second = instr.split('split_at', remove_unused_parameters=True)
        try:
            compile_and_run(second, '-n 1000 ang=10', run=True)
        except RuntimeError as e:
            logger.error(f'Failed to compile instrument: {e}')
            self.fail(f'Failed to compile instrument {e}')

    @mcpl_compiled
    def test_mcpl_split_run(self):
        # Adapted from Test_MCPL_*.instr in ${MCCODE}/mcstas-comps/examples
        from mccode_antlr.instr import Instr
        from mccode_antlr.common import ComponentParameter, Expr
        from mccode_antlr.compiler.c import compile_instrument, run_compiled_instrument, CBinaryTarget
        from mccode_antlr.translators.target import MCSTAS_GENERATOR
        from mccode_antlr.loader import read_mccode_dat, parse_mcstas_instr
        from tempfile import TemporaryDirectory
        from os import R_OK, access
        from pathlib import Path
        from random import randint
        from numpy import allclose

        instr_source = """
        DEFINE INSTRUMENT Test_MCPL_output()
        USERVARS %{ double flag; %}
        TRACE
        COMPONENT Origin = Progress_bar() AT (0, 0, 0) ABSOLUTE
        COMPONENT sa = Source_Maxwell_3( /* flux in n/s/cm2/st/AA */
                                        Lmin=1, Lmax=11, dist=1, focus_xw=0.1, focus_yh=0.1, xwidth=0.01, yheight=0.01,
                                        T1=216.8,I1=1.24e+13,T2=33.9,I2=1.02e+13, T3=16.7 ,I3=3.0423e+12)
        AT (0,0,0) ABSOLUTE
        EXTEND %{
            t=1e-3*rand01();
            flag=(double)mcget_run_num();
        %}
        COMPONENT split_point = Arm() AT(0,0,0) RELATIVE PREVIOUS
        COMPONENT m1 = Monitor_nD(filename="o1", xwidth=0.2, yheight=0.2,
                                  options="lambda limits=[1 11] bins=100 parallel", bins=40)
        AT (0,0,0) ABSOLUTE
        COMPONENT m2 = Monitor_nD(filename="o2", xwidth=0.2, yheight=0.2, options="x y, parallel", bins=40)
        AT (0,0,0) ABSOLUTE
        COMPONENT m3 = Monitor_nD(filename="o3", xwidth=0.2, yheight=0.2, options="t limits=[0 1e-3]parallel", bins=40)
        AT (0,0,0) ABSOLUTE
        COMPONENT m4 = Monitor_nD(filename="o4", xwidth=0.2, yheight=0.2, options="E limits=[0 82] parallel", bins=40)
        AT (0,0,0) ABSOLUTE
        END
        """
        instr = parse_mcstas_instr(instr_source)
        self.assertEqual(instr.name, 'Test_MCPL_output')

        flag_params = tuple(ComponentParameter(k, Expr.str(v)) for k, v in [
            ('userflag', '"flag"'), ('userflagcomment', '"Neutron Id"')])
        before, after = instr.mcpl_split('split_point', filename='test_mcpl_split', output_parameters=flag_params)
        self.assertEqual(before.name, 'Test_MCPL_output_first')
        self.assertEqual(after.name, 'Test_MCPL_output_second')
        self.assertTrue(isinstance(before, Instr))
        self.assertTrue(isinstance(after, Instr))
        self.assertEqual(len(before.components), 3)
        self.assertEqual(len(after.components), 5)
        for i, name in enumerate(('Origin', 'sa', 'split_point')):
            self.assertEqual(before.components[i].name, name)
        for i, name in enumerate(('split_point', 'm1', 'm2', 'm3', 'm4')):
            self.assertEqual(after.components[i].name, name)

        # If we compile and run the parsed instrument, or the split instrument parts, we should get the same results
        target = CBinaryTarget(mpi=False, acc=False, count=1, nexus=False)
        config = dict(default_main=True, enable_trace=False, portable=False, include_runtime=True,
                      embed_instrument_file=False, verbose=False)
        with TemporaryDirectory() as directory:
            expected_binaries = [Path(directory).joinpath(f'{x.name}.out') for x in (instr, before, after)]
            for obj in (instr, before, after):
                try:
                    compile_instrument(obj, target, directory, generator=MCSTAS_GENERATOR, config=config)
                except RuntimeError as e:
                    logger.error(f'Failed to compile instrument: {e}')
                    raise e

            for binary in expected_binaries:
                self.assertTrue(binary.exists())
                self.assertTrue(binary.is_file())
                self.assertTrue(access(binary, R_OK))

            # Run the instrument and check that the output is the same
            seed = randint(1000, 2**32 - 1)
            common = f'--seed {seed} -n 10000'
            run_compiled_instrument(expected_binaries[0], target, f"--dir {directory}/instr {common}")
            mcpl_file = Path(directory).joinpath('instr')
            run_compiled_instrument(expected_binaries[1], target, f"--dir {directory}/before {common} mcpl_filename={mcpl_file}")
            # Breaking change in development McStas -- MCPL_output appends the extension even if it's already there
            # So do not specify it, then add it here.
            if not mcpl_file.exists() or not mcpl_file.is_file():
                mcpl_file = Path(directory).joinpath('instr.mcpl')
            # Depending on how MCPL was built, it might have gzipped the output file
            if not mcpl_file.exists() or not mcpl_file.is_file():
                mcpl_file = Path(directory).joinpath('instr.mcpl.gz')
            self.assertTrue(mcpl_file.exists())
            try:
                run_compiled_instrument(expected_binaries[2], target,
                                        f"--dir {directory}/after {common} mcpl_filename={mcpl_file}")
            except RuntimeError as ex:
                self.fail(str(ex))

            # Now check that the instr and after outputs are the same
            instr_files = list(Path(directory).joinpath('instr').glob('o*'))
            after_files = list(Path(directory).joinpath('after').glob('o*'))
            self.assertEqual(len(instr_files), len(after_files))
            for file in instr_files:
                self.assertEqual(len([x for x in after_files if x.name == file.name]), 1)
            for file in after_files:
                self.assertEqual(len([x for x in instr_files if x.name == file.name]), 1)
            instr_files = [[x for x in instr_files if x.name == file.name][0] for file in after_files]
            for instr_file, after_file in zip(instr_files, after_files):
                # The files must be the same except for any header information:
                instr_data = read_mccode_dat(instr_file)
                after_data = read_mccode_dat(after_file)
                self.assertTrue(allclose(instr_data.data, after_data.data))


    def test_assemble_a3_rotation(self):
        from mccode_antlr.assembler import Assembler
        from mccode_antlr.reader import MCSTAS_REGISTRY
        a = Assembler('test_a3_angle', registries=[MCSTAS_REGISTRY])
        a.parameter('phase/"degree" = 0')
        a.parameter('a3/"degree" = 0')
        a.component('origin', 'Progress_bar')
        a.component('source', 'ESS_butterfly', at=[(0, 0, 0), 'origin'],
                    parameters={'sector': '"W"', 'beamline': 4, 'yheight': 0.03, 'cold_frac': 0.5,
                                'dist': 1, 'focus_xw': 0.01, 'focus_yh': 0.01, 'c_performance': 1.0,
                                't_performance': 1.0, 'Lmin': 1, 'Lmax': 2, 'tmax_multiplier': 1.5,
                                'n_pulses': 1, 'acc_power': 2.0})
        a.component('guide', 'Guide_gravity', at=[(0, 0, 1), 'origin'],
                    parameters={'w1': 0.01, 'w2': 0.05, 'h1': 0.01, 'h2': 0.05, 'l': 8.0, 'm': 3.5, 'G':-9.82})
        a.component('guide_end', 'Arm', at=[(0, 0, 8.0), 'guide'])
        a.component('chopper', 'DiskChopper', at=([0, 0, 0.01], 'guide_end'),
                    parameters={'radius':0.35, 'nu': 14, 'phase': 'phase', 'theta_0': 115.0})
        a.component('rotate', 'Arm', at=([0, 0, 0], 'chopper'), rotate=([0, 'a3', 0], 'chopper'))
        a.component('aperture', 'Slit', at=([0, 0, 1], 'rotate'),
                    parameters={'xwidth': 0.05, 'yheight': 0.05})
        a.component('sample', 'Incoherent', at=([0, 0, 1], 'aperture'),
                    parameters={'radius': 0.005, 'yheight': 0.02, 'thickness': 0.001, 'focus_ah': 2.0,
                                'focus_aw': 2.0, 'target_x': 1.0, 'target_y': 0.0, 'target_z': 0.0,
                                'Etrans': 0.0, 'deltaE': 0.1})
        instr = a.instrument
        try:
            compile_and_run(instr, '-n 1000 a3=10', run=True)
        except RuntimeError as e:
            logger.error(f'Failed to compile instrument: {e}')
            self.fail(f'Failed to compile instrument {e}')